"""Provide tests for marker class to calculate cross-parcellation FC."""

# Authors: Leonard Sasse <l.sasse@fz-juelich.de>
#          Kaustubh R. Patil <k.patil@fz-juelich.de>
# License: AGPL

from pathlib import Path

import pytest
from nilearn import image

from junifer.markers.crossparcellation_functional_connectivity import (
    CrossParcellationFC,
)
from junifer.storage import SQLiteFeatureStorage
from junifer.testing.datagrabbers import SPMAuditoryTestingDatagrabber


parcellation_ONE = "Schaefer100x17"
parcellation_TWO = "Schaefer200x17"


def test_compute() -> None:
    """Test CrossParcellationFC compute()."""

    with SPMAuditoryTestingDatagrabber() as dg:
        out = dg["sub001"]
        niimg = image.load_img(str(out["BOLD"]["path"].absolute()))
        input_dict = {
            "BOLD": {
                "data": niimg,
                "path": out["BOLD"]["path"],
                "meta": {"element": "sub001"},
            }
        }

        crossparcellation = CrossParcellationFC(
            parcellation_one=parcellation_ONE,
            parcellation_two=parcellation_TWO,
            correlation_method="spearman",
        )
        out = crossparcellation.compute(input_dict["BOLD"])
        assert out["data"].shape == (200, 100)
        assert len(out["col_names"]) == 100
        assert len(out["row_names"]) == 200


def test_store(tmp_path: Path) -> None:
    """Test CrossParcellationFC store().

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """

    with SPMAuditoryTestingDatagrabber() as dg:
        input_dict = dg["sub001"]
        niimg = image.load_img(str(input_dict["BOLD"]["path"].absolute()))

        input_dict["BOLD"]["data"] = niimg

        crossparcellation = CrossParcellationFC(
            parcellation_one=parcellation_ONE,
            parcellation_two=parcellation_TWO,
            correlation_method="spearman",
        )
        uri = tmp_path / "test_crossparcellation.sqlite"
        storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
        crossparcellation.fit_transform(input_dict, storage=storage)
        features = storage.list_features()
        assert any(
            x["name"] == "BOLD_CrossParcellationFC"
            for x in features.values()
        )


def test_get_output_type() -> None:
    """Test CrossParcellationFC get_output_type()."""

    crossparcellation = CrossParcellationFC(
        parcellation_one=parcellation_ONE, parcellation_two=parcellation_TWO
    )
    input_ = "BOLD"
    output = crossparcellation.get_output_type(input_)
    assert output == "matrix"


def test_init_() -> None:
    """Test CrossParcellationFC init()."""
    with pytest.raises(ValueError, match="must be different"):
        CrossParcellationFC(
            parcellation_one="a",
            parcellation_two="a",
            correlation_method="pearson",
        )



"""Provide tests for coordinates."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from junifer.data.coordinates import (
    list_coordinates,
    load_coordinates,
    register_coordinates,
)


def test_register_coordinates_built_in_check() -> None:
    """Test coordinates registration check for built-in coordinates."""
    with pytest.raises(ValueError, match=r"built-in"):
        register_coordinates(
            name="DMNBuckner",
            coordinates=np.zeros(2),
            voi_names=["1", "2"],
            overwrite=True,
        )


def test_register_coordinates_overwrite() -> None:
    """Test coordinates registration check for overwriting."""
    register_coordinates(
        name="MyList",
        coordinates=np.zeros((2, 3)),
        voi_names=["roi1", "roi2"],
        overwrite=True,
    )
    with pytest.raises(ValueError, match=r"already registered"):
        register_coordinates(
            name="MyList",
            coordinates=np.ones((2, 3)),
            voi_names=["roi2", "roi3"],
        )

    register_coordinates(
        name="MyList",
        coordinates=np.ones((2, 3)),
        voi_names=["roi2", "roi3"],
        overwrite=True,
    )

    coord, names = load_coordinates("MyList")
    assert_array_equal(coord, np.ones((2, 3)))
    assert names == ["roi2", "roi3"]


def test_register_coordinates_valid_input() -> None:
    """Test coordinates registration check for valid input."""
    with pytest.raises(ValueError, match=r"numpy.ndarray"):
        register_coordinates(
            name="MyList",
            coordinates=[1, 2],
            voi_names=["roi1", "roi2"],
            overwrite=True,
        )
    with pytest.raises(ValueError, match=r"2D array"):
        register_coordinates(
            name="MyList",
            coordinates=np.zeros((2, 3, 4)),
            voi_names=["roi1", "roi2"],
            overwrite=True,
        )

    with pytest.raises(ValueError, match=r"3 values"):
        register_coordinates(
            name="MyList",
            coordinates=np.zeros((2, 4)),
            voi_names=["roi1", "roi2"],
            overwrite=True,
        )
    with pytest.raises(ValueError, match=r"voi_names"):
        register_coordinates(
            name="MyList",
            coordinates=np.zeros((2, 3)),
            voi_names=["roi1", "roi2", "roi3"],
            overwrite=True,
        )


def test_list_coordinates() -> None:
    """Test listing of available coordinates."""
    available_coordinates = list_coordinates()
    assert "DMNBuckner" in available_coordinates
    assert "MultiTask" in available_coordinates
    assert "VigAtt" in available_coordinates
    assert "WM" in available_coordinates


def test_load_coordinates() -> None:
    """Test loading coordinates from file."""
    coord, names = load_coordinates("DMNBuckner")
    assert coord.shape == (6, 3)  # type: ignore
    assert names == ["PCC", "MPFC", "lAG", "rAG", "lHF", "rHF"]


def test_load_coordinates_nonexisting() -> None:
    """Test loading coordinates that not exist."""
    with pytest.raises(ValueError, match=r"not found"):
        load_coordinates("NonExisting")



"""Provide functions for cli."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import shutil
import subprocess
import textwrap
import typing
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import yaml

from ..datagrabber.base import BaseDataGrabber
from ..markers.base import BaseMarker
from ..markers.collection import MarkerCollection
from ..pipeline.registry import build
from ..storage.base import BaseFeatureStorage
from ..utils import logger, raise_error
from ..utils.fs import make_executable


def _get_datagrabber(datagrabber_config: Dict) -> BaseDataGrabber:
    """Get datagrabber.

    Parameters
    ----------
    datagrabber_config : dict
        The config to get the datagrabber using.

    Returns
    -------
    dict
        The datagrabber.

    """
    datagrabber_params = datagrabber_config.copy()
    datagrabber_kind = datagrabber_params.pop("kind")
    datagrabber = build(
        step="datagrabber",
        name=datagrabber_kind,
        baseclass=BaseDataGrabber,
        init_params=datagrabber_params,
    )
    datagrabber = typing.cast(BaseDataGrabber, datagrabber)
    return datagrabber


def run(
    workdir: Union[str, Path],
    datagrabber: Dict,
    markers: List[Dict],
    storage: Dict,
    elements: Union[str, List[Union[str, Tuple]], Tuple, None] = None,
) -> None:
    """Run the pipeline on the selected element.

    Parameters
    ----------
    workdir : str or pathlib.Path
        Directory where the pipeline will be executed.
    datagrabber : dict
        Datagrabber to use. Must have a key ``kind`` with the kind of
        datagrabber to use. All other keys are passed to the datagrabber
        init function.
    markers : list of dict
        List of markers to extract. Each marker is a dict with at least two
        keys: ``name`` and ``kind``. The ``name`` key is used to name the
        output marker. The ``kind`` key is used to specify the kind of marker
        to extract. The rest of the keys are used to pass parameters to the
        marker calculation.
    storage : dict
        Storage to use. Must have a key ``kind`` with the kind of
        storage to use. All other keys are passed to the storage
        init function.
    elements : str or tuple or list of str or tuple, optional
        Element(s) to process. Will be used to index the datagrabber
        (default None).

    """
    # Convert str to Path
    if isinstance(workdir, str):
        workdir = Path(workdir)
    if not isinstance(elements, list) and elements is not None:
        elements = [elements]
    # Get datagrabber to use
    datagrabber_object = _get_datagrabber(datagrabber)
    # Copy to avoid changing the original dict
    _markers = [x.copy() for x in markers]
    built_markers = []
    for t_marker in _markers:
        kind = t_marker.pop("kind")
        t_m = build(
            step="marker",
            name=kind,
            baseclass=BaseMarker,
            init_params=t_marker,
        )
        built_markers.append(t_m)
    # Get storage engine to use
    storage_params = storage.copy()
    storage_kind = storage_params.pop("kind")
    if "single_output" not in storage_params:
        storage_params["single_output"] = False
    storage_object = build(
        step="storage",
        name=storage_kind,
        baseclass=BaseFeatureStorage,
        init_params=storage_params,
    )
    storage_object = typing.cast(BaseFeatureStorage, storage_object)
    # Create new marker collection
    mc = MarkerCollection(markers=built_markers, storage=storage_object)
    # Fit elements
    with datagrabber_object:
        if elements is not None:
            for t_element in elements:
                mc.fit(datagrabber_object[t_element])
        else:
            for t_element in datagrabber_object:
                mc.fit(datagrabber_object[t_element])


def collect(storage: Dict) -> None:
    """Collect and store data.

    Parameters
    ----------
    storage : dict
        Storage to use. Must have a key ``kind`` with the kind of
        storage to use. All other keys are passed to the storage
        init function.

    """
    storage_params = storage.copy()
    storage_kind = storage_params.pop("kind")
    logger.info(f"Collecting data using {storage_kind}")
    logger.debug(f"\tStorage params: {storage_params}")
    if "single_output" not in storage_params:
        storage_params["single_output"] = False
    storage_object = build(
        step="storage",
        name=storage_kind,
        baseclass=BaseFeatureStorage,
        init_params=storage_params,
    )
    storage_object = typing.cast(BaseFeatureStorage, storage_object)
    logger.debug("Running storage.collect()")
    storage_object.collect()
    logger.info("Collect done")


def queue(
    config: Dict,
    kind: str,
    jobname: str = "junifer_job",
    overwrite: bool = False,
    elements: Union[str, List[Union[str, Tuple]], Tuple, None] = None,
    **kwargs: Union[str, int, bool, Dict, Tuple, List],
) -> None:
    """Queue a job to be executed later.

    Parameters
    ----------
    config : dict
        The configuration to be used for queueing the job.
    kind : {"HTCondor", "SLURM"}
        The kind of job queue system to use.
    jobname : str, optional
        The name of the job (default "junifer_job").
    overwrite : bool, optional
        Whether to overwrite if job directory already exists (default False).
    elements : str or tuple or list of str or tuple, optional
        Element(s) to process. Will be used to index the datagrabber
        (default None).
    **kwargs : dict
        The keyword arguments to pass to the job queue system.

    Raises
    ------
    ValueError
        If the value of ``kind`` is invalid.

    """
    # Create a folder within the CWD to store the job files / config
    cwd = Path.cwd()
    jobdir = cwd / "junifer_jobs" / jobname
    logger.info(f"Creating job in {str(jobdir.absolute())}")
    if jobdir.exists():
        if not overwrite:
            raise_error(
                f"Job folder for {jobname} already exists. "
                "This error is raised to prevent overwriting job files "
                "that might be scheduled but not yet executed. "
                f"Either delete the directory {str(jobdir.absolute())} "
                "or set overwrite=True."
            )
        else:
            logger.info(
                f"Deleting existing job directory at {str(jobdir.absolute())}"
            )
            shutil.rmtree(jobdir)
    jobdir.mkdir(exist_ok=True, parents=True)

    if "with" in config:
        to_load = config["with"]
        # If there is a list of files to load, copy and remove the path
        # component
        fixed_load = []
        if not isinstance(to_load, list):
            to_load = [to_load]
        for item in to_load:
            if item.endswith(".py"):
                logger.debug(f"Copying {item} to jobdir ({jobdir.absolute()})")
                shutil.copy(item, jobdir)
                fixed_load.append(Path(item).name)
            else:
                fixed_load.append(item)
        config["with"] = fixed_load

    yaml_config = jobdir / "config.yaml"
    logger.info(f"Writing YAML config to {str(yaml_config.absolute())}")
    with open(yaml_config, "w") as f:
        f.write(yaml.dump(config))

    # Get list of elements
    if elements is None:
        if "elements" in config:
            elements = config["elements"]
        else:
            # If no elements are specified, use all elements from the
            # datagrabber
            datagrabber = _get_datagrabber(config["datagrabber"])
            with datagrabber as dg:
                elements = dg.get_elements()

    # TODO: Fix typing of elements
    if not isinstance(elements, list):
        elements = [elements]  # type: ignore

    typing.cast(List[Union[str, Tuple]], elements)

    if kind == "HTCondor":
        _queue_condor(
            jobname=jobname,
            jobdir=jobdir,
            yaml_config=yaml_config,
            elements=elements,  # type: ignore
            config=config,
            **kwargs,
        )
    elif kind == "SLURM":
        _queue_slurm(
            jobname=jobname,
            jobdir=jobdir,
            yaml_config=yaml_config,
            elements=elements,  # type: ignore
            config=config,
            **kwargs,
        )
    else:
        raise_error(f"Unknown queue kind: {kind}")

    logger.info("Queue done")


def _queue_condor(
    jobname: str,
    jobdir: Path,
    yaml_config: Path,
    elements: List[Union[str, Tuple]],
    config: Dict,
    env: Optional[Dict[str, str]] = None,
    mem: str = "8G",
    cpus: int = 1,
    disk: str = "1G",
    extra_preamble: str = "",
    verbose: str = "info",
    collect: bool = True,
    submit: bool = False,
) -> None:
    """Submit job to HTCondor.

    Parameters
    ----------
    jobname : str
        The name of the job.
    jobdir : pathlib.Path
        The path to the job directory.
    yaml_config : pathlib.Path
        The path to the YAML config file.
    elements : list of str or tuple
        Element(s) to process. Will be used to index the datagrabber.
    config : dict
        The configuration to be used for queueing the job.
    env : dict, optional
        The environment variables passed as dictionary (default None).
    mem : str, optional
        The size of memory (RAM) to use (default "8G").
    cpus : int, optional
        The number of CPU cores to use (default 1).
    disk : str, optional
        The size of disk (HDD or SSD) to use (default "1G").
    extra_preamble : str, optional
        Extra commands to pass to HTCondor (default "").
    verbose : str, optional
        The level of verbosity (default "info").
    collect : bool, optional
        Whether to submit "collect" task for junifer (default True).
    submit : bool, optional
        Whether to submit the jobs. In any case, .dag files will be created
        for submission (default False).

    Raises
    ------
    ValueError
        If the value of `env` is invalid.

    """
    logger.debug("Creating HTCondor job")
    run_junifer_args = (
        f"run {str(yaml_config.absolute())} "
        f"--verbose {verbose} --element $(element)"
    )
    collect_junifer_args = (
        f"collect {str(yaml_config.absolute())} --verbose {verbose} "
    )

    # Set up the env_name, executable and arguments according to the
    # environment type
    if env is None:
        env = {"kind": "local"}
    if env["kind"] == "conda":
        env_name = env["name"]
        executable = "run_conda.sh"
        arguments = f"{env_name} junifer"
        exec_path = jobdir / executable
        logger.info(f"Copying {executable} to {str(exec_path.absolute())}")
        shutil.copy(Path(__file__).parent / "res" / executable, exec_path)
        make_executable(exec_path)
    elif env["kind"] == "venv":
        env_name = env["name"]
        executable = "run_venv.sh"
        arguments = f"{env_name} junifer"
        # TODO: Copy run_venv.sh to jobdir
    elif env["kind"] == "local":
        executable = "junifer"
        arguments = ""
    else:
        raise ValueError(f'Unknown env kind: {env["kind"]}')

    # Create log directory
    log_dir = jobdir / "logs"
    log_dir.mkdir(exist_ok=True, parents=True)

    # Add preamble data
    run_preamble = f"""
        # The environment
        universe = vanilla
        getenv = True

        # Resources
        request_cpus = {cpus}
        request_memory = {mem}
        request_disk = {disk}

        # Executable
        initial_dir = {str(jobdir.absolute())}
        executable = $(initial_dir)/{executable}
        transfer_executable = False

        arguments = {arguments} {run_junifer_args}

        {extra_preamble}

        # Logs
        log = {str(log_dir.absolute())}/junifer_run_$(log_element).log
        output = {str(log_dir.absolute())}/junifer_run_$(log_element).out
        error = {str(log_dir.absolute())}/junifer_run_$(log_element).err
        """

    submit_run_fname = jobdir / f"run_{jobname}.submit"
    submit_collect_fname = jobdir / f"collect_{jobname}.submit"
    dag_fname = jobdir / f"{jobname}.dag"

    # Write to run submit files
    with open(submit_run_fname, "w") as submit_file:
        submit_file.write(textwrap.dedent(run_preamble))
        submit_file.write("queue\n")

    collect_preamble = f"""
        # The environment
        universe = vanilla
        getenv = True

        # Resources
        request_cpus = {cpus}
        request_memory = {mem}
        request_disk = {disk}

        # Executable
        initial_dir = {str(jobdir.absolute())}
        executable = $(initial_dir)/{executable}
        transfer_executable = False

        arguments = {arguments} {collect_junifer_args}

        {extra_preamble}

        # Logs
        log = {str(log_dir.absolute())}/junifer_collect.log
        output = {str(log_dir.absolute())}/junifer_collect.out
        error = {str(log_dir.absolute())}/junifer_collect.err
        """

    # Now create the collect submit file
    with open(submit_collect_fname, "w") as submit_file:
        submit_file.write(textwrap.dedent(collect_preamble))
        submit_file.write("queue\n")

    with open(dag_fname, "w") as dag_file:
        # Get all subject and session names from file list
        for i_job, t_elem in enumerate(elements):
            str_elem = (
                ",".join(t_elem) if isinstance(t_elem, tuple) else t_elem
            )
            log_elem = (
                "_".join(t_elem) if isinstance(t_elem, tuple) else t_elem
            )
            dag_file.write(f"JOB run{i_job} {submit_run_fname}\n")
            dag_file.write(
                f'VARS run{i_job} element="{str_elem}" '
                f'log_element="{log_elem}"\n\n'
            )
        if collect is True:
            dag_file.write(f"JOB collect {submit_collect_fname}\n")
            dag_file.write("PARENT ")
            for i_job, _t_elem in enumerate(elements):
                dag_file.write(f"run{i_job} ")
            dag_file.write("CHILD collect\n\n")

    # Submit job(s)
    if submit is True:
        logger.info("Submitting HTCondor job")
        subprocess.run(["condor_submit_dag", dag_fname])
        logger.info("HTCondor job submitted")
    else:
        cmd = f"condor_submit_dag {str(dag_fname.absolute())}"
        logger.info(
            f"HTCondor job files created, to submit the job, run `{cmd}`"
        )


def _queue_slurm(
    jobname: str,
    jobdir: Path,
    yaml_config: Path,
    elements: List[Union[str, Tuple]],
    config: Dict,
) -> None:
    """Submit job to SLURM.

    Parameters
    ----------
    jobname : str
        The name of the job.
    jobdir : pathlib.Path
        The path to the job directory.
    yaml_config : pathlib.Path
        The path to the YAML config file.
    elements : str or tuple or list[str or tuple], optional
        Element(s) to process. Will be used to index the datagrabber
        (default None).
    config : dict
        The configuration to be used for queueing the job.
    """
    pass
    # logger.debug("Creating SLURM job")
    # run_junifer_args = (
    #     f"run {str(yaml_config.absolute())} "
    #     f"--verbose {verbose} --element $(element)"
    # )
    # collect_junifer_args = \
    #     f"collect {str(yaml_config.absolute())} --verbose {verbose} "

    # # Set up the env_name, executable and arguments according to the
    # # environment type
    # if env is None:
    #     env = {
    #         "kind": "local",
    #     }
    # if env["kind"] == "conda":
    #     env_name = env["name"]
    #     executable = "run_conda.sh"
    #     arguments = f"{env_name} junifer"
    #     # TODO: Copy run_conda.sh to jobdir
    #     exec_path = jobdir / executable
    #     shutil.copy(Path(__file__).parent / "res" / executable, exec_path)
    #     make_executable(exec_path)
    # elif env["kind"] == "venv":
    #     env_name = env["name"]
    #     executable = "run_venv.sh"
    #     arguments = f"{env_name} junifer"
    #     # TODO: Copy run_venv.sh to jobdir
    # elif env["kind"] == "local":
    #     executable = "junifer"
    #     arguments = ""
    # else:
    #     raise ValueError(f"Unknown env kind: {env['kind']}")

    # # Create log directory
    # log_dir = jobdir / 'logs'
    # log_dir.mkdir(exist_ok=True, parents=True)

    # # Add preamble data
    # run_preamble = f"""
    #     #!/bin/bash

    #     #SBATCH --job-name={}
    #     #SBATCH --account={}
    #     #SBATCH --partition={}
    #     #SBATCH --time={}
    #     #SBATCH --ntasks={}
    #     #SBATCH --cpus-per-task={cpus}
    #     #SBATCH --mem-per-cpu={mem}
    #     #SBATCH --mail-type={}
    #     #SBATCH --mail-user={}
    #     #SBATCH --output={}
    #     #SBATCH --error={}

    #     # Executable
    #     initial_dir = {str(jobdir.absolute())}
    #     executable = $(initial_dir)/{executable}
    #     transfer_executable = False

    #     arguments = {arguments} {run_junifer_args}

    #     {extra_preamble}

    #     # Logs
    #     log = {str(log_dir.absolute())}/junifer_run_$(element).log
    #     output = {str(log_dir.absolute())}/junifer_run_$(element).out
    #     error = {str(log_dir.absolute())}/junifer_run_$(element).err
    #     """

    # submit_run_fname = jobdir / f'run_{jobname}.sh'
    # submit_collect_fname = jobdir / f'collect_{jobname}.sh'

    # # Write to run submit files
    # with open(submit_run_fname, 'w') as submit_file:
    #     submit_file.write(run_preamble)
    #     submit_file.write('queue\n')

    # collect_preamble = f"""
    #     # The environment
    #     universe = vanilla
    #     getenv = True

    #     # Resources
    #     request_cpus = {cpus}
    #     request_memory = {mem}
    #     request_disk = {disk}

    #     # Executable
    #     initial_dir = {str(jobdir.absolute())}
    #     executable = $(initial_dir)/{executable}
    #     transfer_executable = False

    #     arguments = {arguments} {collect_junifer_args}

    #     {extra_preamble}

    #     # Logs
    #     log = {str(log_dir.absolute())}/junifer_collect.log
    #     output = {str(log_dir.absolute())}/junifer_collect.out
    #     error = {str(log_dir.absolute())}/junifer_collect.err
    #     """

    # # Now create the collect submit file
    # with open(submit_collect_fname, 'w') as submit_file:
    #     submit_file.write(collect_preamble)  # Eval preamble here
    #     submit_file.write('queue\n')

    # with open(dag_fname, 'w') as dag_file:
    #     # Get all subject and session names from file list
    #     for i_job, t_elem in enumerate(elements):
    #         dag_file.write(f'JOB run{i_job} {submit_run_fname}\n')
    #         dag_file.write(f'VARS run{i_job} element="{t_elem}"\n\n')
    #     if collect is True:
    #         dag_file.write(f'JOB collect {submit_collect_fname}\n')
    #         dag_file.write('PARENT ')
    #         for i_job, _t_elem in enumerate(elements):
    #             dag_file.write(f'run{i_job} ')
    #         dag_file.write('CHILD collect\n\n')

    # # Submit job(s)
    # if submit is True:
    #     logger.info('Submitting SLURM job')
    #     subprocess.run(['condor_submit_dag', dag_fname])
    #     logger.info('HTCondor SLURM submitted')
    # else:
    #     cmd = f"condor_submit_dag {str(dag_fname.absolute())}"
    #     logger.info(
    #         f"SLURM job files created, to submit the job, run `{cmd}`"
    #     )



"""Provide abstract base class for feature storage via pandas."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import json
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from .base import BaseFeatureStorage


class PandasBaseFeatureStorage(BaseFeatureStorage):
    """Abstract base class for feature storage via pandas.

    For every interface that is required, one needs to provide a concrete
    implementation of this abstract class.

    Parameters
    ----------
    uri : str or pathlib.Path
        The path to the storage.
    single_output : bool, optional
        Whether to have single output (default True).
    **kwargs
        Keyword arguments passed to superclass.

    See Also
    --------
    BaseFeatureStorage : The base class for feature storage.

    """

    def __init__(
        self, uri: Union[str, Path], single_output: bool = True, **kwargs
    ) -> None:
        super().__init__(uri=uri, single_output=single_output, **kwargs)

    def get_valid_inputs(self) -> List[str]:
        """Get valid storage types for input.

        Returns
        -------
        list of str
            The list of storage types that can be used as input for this "
            "storage.
        """
        return ["matrix", "table", "timeseries"]

    def _meta_row(self, meta: Dict, meta_md5: str) -> pd.DataFrame:
        """Convert the metadata to a pandas DataFrame.

        Parameters
        ----------
        meta : dict
            The metadata as a dictionary.
        meta_md5 : str
            The MD5 hash of the metadata.

        Returns
        -------
        pandas.DataFrame

        """
        data_df = {}
        for k, v in meta.items():
            data_df[k] = json.dumps(v, sort_keys=True)
        df = pd.DataFrame(data_df, index=[meta_md5])
        df.index.name = "meta_md5"
        return df

    @staticmethod
    def element_to_index(
        element: Dict, n_rows: int = 1, rows_col_name: Optional[str] = None
    ) -> pd.MultiIndex:
        """Convert the element metadata to index.

        Parameters
        ----------
        element : dict
            The element as a dictionary.
        n_rows : int, optional
            Number of rows to create (default 1).
        rows_col_name: str, optional
            The column name to use in case `n_rows` > 1. If None and
            n_rows > 1, the name will be "idx" (default None).

        Returns
        -------
        pandas.MultiIndex
            The index of the dataframe to store.

        Raises
        ------
        ValueError
            If `meta` does not contain the key "element".

        """
        # Check rows_col_name
        if rows_col_name is None:
            rows_col_name = "idx"
        elem_idx: Dict[Any, Any] = {
            k: [v] * n_rows for k, v in element.items()
        }
        elem_idx[rows_col_name] = np.arange(n_rows)
        # Create index
        index = pd.MultiIndex.from_frame(
            pd.DataFrame(elem_idx, index=range(n_rows))
        )
        return index

    def store_df(
        self, meta_md5: str, element: Dict, df: Union[pd.DataFrame, pd.Series]
    ) -> None:
        """Implement pandas DataFrame storing.

        Parameters
        ----------
        df : pandas.DataFrame or pandas.Series
            The pandas DataFrame or Series to store.
        meta : dict
            The metadata as a dictionary.

        Raises
        ------
        ValueError
            If the dataframe index has items that are not in the index
            generated from the metadata.

        """
        raise NotImplementedError("Implement in subclass.")

    def _store_2d(
        self,
        meta_md5: str,
        element: Dict,
        data: Union[np.ndarray, List],
        columns: Optional[Iterable[str]] = None,
        rows_col_name: Optional[str] = None,
    ) -> None:
        """Store 2D dataframe.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray or List
            The data to store.
        columns : list or tuple of str, optional
            The columns (default None).
        rows_col_name : str, optional
            The column name to use in case number of rows greater than 1.
            If None and number of rows greater than 1, then the name will be
            "index" (default None).

        """
        n_rows = len(data)
        # Convert element metadata to index
        idx = self.element_to_index(
            element=element, n_rows=n_rows, rows_col_name=rows_col_name
        )
        # Prepare new dataframe
        data_df = pd.DataFrame(  # type: ignore
            data, columns=columns, index=idx  # type: ignore
        )
        # Store dataframe
        self.store_df(meta_md5=meta_md5, element=element, df=data_df)

    def store_table(
        self,
        meta_md5: str,
        element: Dict,
        data: Union[np.ndarray, List],
        columns: Optional[Iterable[str]] = None,
        rows_col_name: Optional[str] = None,
    ) -> None:
        """Implement table storing.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray or List
            The table data to store.
        columns : list or tuple of str, optional
            The columns (default None).
        rows_col_name : str, optional
            The column name to use in case number of rows greater than 1.
            If None and number of rows greater than 1, then the name will be
            "index" (default None).
        """
        self._store_2d(
            meta_md5=meta_md5,
            element=element,
            data=data,
            columns=columns,
            rows_col_name=rows_col_name,
        )

    def store_timeseries(
        self,
        meta_md5: str,
        element: Dict,
        data: np.ndarray,
        columns: Optional[Iterable[str]] = None,
    ) -> None:
        """Implement timeseries storing.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray
            The timeseries data to store.
        columns : list or tuple of str, optional
            The column labels (default None).
        """
        self._store_2d(
            meta_md5=meta_md5,
            element=element,
            data=data,
            columns=columns,
            rows_col_name="timepoint",
        )



"""Provide tests for JuniferNiftiSpheresMasker class."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import warnings

import nibabel
import numpy as np
import pytest
from nilearn._utils import data_gen
from nilearn.image import get_data
from numpy.testing import assert_array_equal

from junifer.external.nilearn import JuniferNiftiSpheresMasker


"""
New BSD License

Copyright (c) 2007 - 2022 The nilearn developers.
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of the nilearn developers nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""


def test_seed_extraction() -> None:
    """Test seed extraction."""
    data = np.random.RandomState(42).random_sample((3, 3, 3, 5))
    img = nibabel.Nifti1Image(data, np.eye(4))
    masker = JuniferNiftiSpheresMasker(seeds=[(1, 1, 1)])
    # Test the fit
    masker.fit()
    # Test the transform
    s = masker.transform(img)
    assert_array_equal(s[:, 0], data[1, 1, 1])


def test_sphere_extraction() -> None:
    """Test sphere extraction."""
    data = np.random.RandomState(42).random_sample((3, 3, 3, 5))
    img = nibabel.Nifti1Image(data, np.eye(4))
    masker = JuniferNiftiSpheresMasker(seeds=[(1, 1, 1)], radius=1)

    # Check attributes defined at fit
    assert not hasattr(masker, "seeds_")
    assert not hasattr(masker, "n_elements_")

    masker.fit()

    # Check attributes defined at fit
    assert hasattr(masker, "seeds_")
    assert hasattr(masker, "n_elements_")
    assert masker.n_elements_ == 1

    # Test the fit
    masker.fit()

    # Test the transform
    s = masker.transform(img)
    mask = np.zeros((3, 3, 3), dtype=bool)
    mask[:, 1, 1] = True
    mask[1, :, 1] = True
    mask[1, 1, :] = True
    assert_array_equal(s[:, 0], np.mean(data[mask], axis=0))

    # Now with a mask
    mask_img = np.zeros((3, 3, 3))
    mask_img[1, :, :] = 1
    mask_img = nibabel.Nifti1Image(mask_img, np.eye(4))
    masker = JuniferNiftiSpheresMasker(
        seeds=[(1, 1, 1)],
        radius=1,
        mask_img=mask_img,
    )
    masker.fit()
    s = masker.transform(img)
    assert_array_equal(
        s[:, 0],
        np.mean(
            data[np.logical_and(mask, get_data(mask_img))],
            axis=0,
        ),
    )


def test_anisotropic_sphere_extraction() -> None:
    """Test anisotropic sphere extraction."""
    data = np.random.RandomState(42).random_sample((3, 3, 3, 5))
    affine = np.eye(4)
    affine[0, 0] = 2
    affine[2, 2] = 2
    img = nibabel.Nifti1Image(data, affine)
    masker = JuniferNiftiSpheresMasker(seeds=[(2, 1, 2)], radius=1)
    # Test the fit
    masker.fit()
    # Test the transform
    s = masker.transform(img)
    mask = np.zeros((3, 3, 3), dtype=bool)
    mask[1, :, 1] = True
    assert_array_equal(s[:, 0], np.mean(data[mask], axis=0))
    # Now with a mask
    mask_img = np.zeros((3, 2, 3))
    mask_img[1, 0, 1] = 1
    affine_2 = affine.copy()
    affine_2[0, 0] = 4
    mask_img = nibabel.Nifti1Image(mask_img, affine=affine_2)
    masker = JuniferNiftiSpheresMasker(
        seeds=[(2, 1, 2)],
        radius=1,
        mask_img=mask_img,
    )

    masker.fit()
    s = masker.transform(img)
    assert_array_equal(s[:, 0], data[1, 0, 1])


def test_errors() -> None:
    """Test errors."""
    masker = JuniferNiftiSpheresMasker(seeds=([1, 2]), radius=0.2)
    with pytest.raises(ValueError, match="Seeds must be a list .+"):
        masker.fit()


def test_nifti_spheres_masker_overlap() -> None:
    """Test overlapping sphere extraction."""
    # Test resampling in NiftiMapsMasker
    affine = np.eye(4)
    shape = (5, 5, 5)

    data = np.random.RandomState(42).random_sample(shape + (5,))
    fmri_img = nibabel.Nifti1Image(data, affine)

    seeds = [(0, 0, 0), (2, 2, 2)]

    overlapping_masker = JuniferNiftiSpheresMasker(
        seeds,
        radius=1,
        allow_overlap=True,
    )
    overlapping_masker.fit_transform(fmri_img)
    overlapping_masker = JuniferNiftiSpheresMasker(
        seeds,
        radius=2,
        allow_overlap=True,
    )
    overlapping_masker.fit_transform(fmri_img)

    noverlapping_masker = JuniferNiftiSpheresMasker(
        seeds,
        radius=1,
        allow_overlap=False,
    )
    noverlapping_masker.fit_transform(fmri_img)
    noverlapping_masker = JuniferNiftiSpheresMasker(
        seeds,
        radius=2,
        allow_overlap=False,
    )
    with pytest.raises(ValueError, match="Overlap detected"):
        noverlapping_masker.fit_transform(fmri_img)


def test_small_radius() -> None:
    """Test sphere extraction with small radius."""
    affine = np.eye(4)
    shape = (3, 3, 3)

    data = np.random.RandomState(42).random_sample(shape)
    mask = np.zeros(shape)
    mask[1, 1, 1] = 1
    mask[2, 2, 2] = 1
    affine = np.eye(4) * 1.2
    seed = (1.4, 1.4, 1.4)

    masker = JuniferNiftiSpheresMasker(
        seeds=[seed],
        radius=0.1,
        mask_img=nibabel.Nifti1Image(mask, affine),
    )
    masker.fit_transform(nibabel.Nifti1Image(data, affine))

    # Test if masking is taken into account
    mask[1, 1, 1] = 0
    mask[1, 1, 0] = 1

    masker = JuniferNiftiSpheresMasker(
        seeds=[seed],
        radius=0.1,
        mask_img=nibabel.Nifti1Image(mask, affine),
    )
    with pytest.raises(ValueError, match="These spheres are empty"):
        masker.fit_transform(nibabel.Nifti1Image(data, affine))

    masker = JuniferNiftiSpheresMasker(
        seeds=[seed],
        radius=1.6,
        mask_img=nibabel.Nifti1Image(mask, affine),
    )
    masker.fit_transform(nibabel.Nifti1Image(data, affine))


def test_is_nifti_spheres_masker_give_nans() -> None:
    """Test NaN interaction with masker."""
    affine = np.eye(4)

    data_with_nans = np.zeros((10, 10, 10), dtype=np.float32)
    data_with_nans[:, :, :] = np.nan

    data_without_nans = np.random.RandomState(42).random_sample((9, 9, 9))
    indices = np.nonzero(data_without_nans)

    # Leaving nans outside of some data
    data_with_nans[indices] = data_without_nans[indices]
    img = nibabel.Nifti1Image(data_with_nans, affine)
    seed = [(7, 7, 7)]

    # Interaction of seed with nans
    masker = JuniferNiftiSpheresMasker(seeds=seed, radius=2.0)
    assert not np.isnan(np.sum(masker.fit_transform(img)))

    mask = np.ones((9, 9, 9))
    mask_img = nibabel.Nifti1Image(mask, affine)
    # When mask_img is provided, the seed interacts within the brain, so no nan
    masker = JuniferNiftiSpheresMasker(
        seeds=seed,
        radius=2.0,
        mask_img=mask_img,
    )
    assert not np.isnan(np.sum(masker.fit_transform(img)))


def test_standardization() -> None:
    """Test standardization."""
    data = np.random.RandomState(42).random_sample((3, 3, 3, 5))
    img = nibabel.Nifti1Image(data, np.eye(4))

    # test zscore
    masker = JuniferNiftiSpheresMasker(seeds=[(1, 1, 1)], standardize="zscore")
    # Test the fit
    s = masker.fit_transform(img)

    np.testing.assert_almost_equal(s.mean(), 0)
    np.testing.assert_almost_equal(s.std(), 1)

    # test psc
    masker = JuniferNiftiSpheresMasker(seeds=[(1, 1, 1)], standardize="psc")
    # Test the fit
    s = masker.fit_transform(img)

    np.testing.assert_almost_equal(s.mean(), 0)
    np.testing.assert_almost_equal(
        s.ravel(),
        data[1, 1, 1] / data[1, 1, 1].mean() * 100 - 100,
    )


def test_nifti_spheres_masker_inverse_transform() -> None:
    """Test inverse transform."""
    data = np.random.RandomState(42).random_sample((3, 3, 3, 5))
    masker = JuniferNiftiSpheresMasker(seeds=[(1, 1, 1)], radius=1)
    with pytest.raises(
        NotImplementedError,
        match="some of which are non-reversible",
    ):
        masker.inverse_transform(data[0, 0, 0, :])


def test_nifti_spheres_masker_io_shapes() -> None:
    """Ensure that masker handles 1D/2D/3D/4D data appropriately.

    transform(4D image) --> 2D output, no warning
    transform(3D image) --> 2D output, DeprecationWarning

    """
    n_regions, n_volumes = 2, 5
    shape_3d = (10, 11, 12)
    shape_4d = (10, 11, 12, n_volumes)
    affine = np.eye(4)

    img_4d, mask_img = data_gen.generate_random_img(
        shape_4d,
        affine=affine,
    )
    img_3d, _ = data_gen.generate_random_img(shape_3d, affine=affine)

    masker = JuniferNiftiSpheresMasker(
        seeds=[(1, 1, 1), (4, 4, 4)],  # number of tuples equal to n_regions
        radius=1,
        mask_img=mask_img,
    )
    masker.fit()

    # DeprecationWarning *should* be raised for 3D inputs
    with pytest.warns(DeprecationWarning, match="Starting in version 0.12"):
        test_data = masker.transform(img_3d)
        assert test_data.shape == (1, n_regions)

    # DeprecationWarning should *not* be raised for 4D inputs
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "error",
            message="Starting in version 0.12",
            category=DeprecationWarning,
        )
        test_data = masker.transform(img_4d)
        assert test_data.shape == (n_volumes, n_regions)



"""Provide test for marker utility functions."""

# Authors: Leonard Sasse <l.sasse@fz-juelich.de>
#          Nicolás Nieto <n.nieto@fz-juelich.de>
#          Sami Hamdan <s.hamdan@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import numpy as np

from junifer.markers.utils import _ets


def test_ets() -> None:
    """Test edge-wise timeseries computing function."""
    bold_ts = np.array(
        [
            [x + 1 for x in range(20)],
            [x + 2 for x in range(20)],
            [x + 3 for x in range(20)],
            [x + 4 for x in range(20)],
        ]
    ).T

    n_time, n_rois = bold_ts.shape
    n_edges = int(n_rois * (n_rois - 1) / 2)

    edge_ts = _ets(bold_ts)
    assert edge_ts.shape == (n_time, n_edges)



"""Provide imports for datagrabber 'aomic' sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL

from .id1000 import DataladAOMICID1000
from .piop1 import DataladAOMICPIOP1
from .piop2 import DataladAOMICPIOP2



# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL
from tempfile import TemporaryDirectory
from pathlib import Path

import datalad.api as dl

dst = 'git@gin.g-node.org:/juaml/datalad-example-bids-ses.git'

with TemporaryDirectory() as tmpdir_name:
    tmpdir = Path(tmpdir_name)
    ds = dl.create(tmpdir)  # type: ignore

    base_dir = tmpdir / 'example_bids_ses'
    base_dir.mkdir()

    for i_sub in range(1, 10):
        t_sub = f'sub-{i_sub:02d}'
        sub_dir = base_dir / t_sub
        sub_dir.mkdir()

        for i_ses in range(1, 4):
            t_ses = f'ses-{i_ses:02d}'
            ses_dir = sub_dir / t_ses
            ses_dir.mkdir()

            for dname in ['anat', 'func']:
                (ses_dir / dname).mkdir()

            fnames = [f'anat/{t_sub}_{t_ses}_T1w.nii.gz']
            if i_ses != 3:  # Session 3 does not have functional data
                fnames.extend([
                    f'func/{t_sub}_{t_ses}_task-rest_bold.nii.gz',
                    f'func/{t_sub}_{t_ses}_task-rest_bold.json'])
            for fname in fnames:
                with open(ses_dir / fname, 'w') as f:
                    f.write('placeholder-{fname}')

    ds.save(recursive=True)
    ds.siblings('add', name='gin', url=dst)
    ds.push(to='gin', force='all')



"""Provide configuration for Sphinx."""

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import sys
from pathlib import Path


curdir = Path(__file__).parent
sys.path.append((curdir / "sphinxext").as_posix())

# -- Project information -----------------------------------------------------

project = "junifer"
copyright = "2022, Authors of junifer"
author = "Fede Raimondo"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # include documentation from docstrings
    "sphinx.ext.autosummary",  # generate autodoc summaries
    "sphinx.ext.doctest",  # test snippets in the documentation
    "sphinx.ext.intersphinx",  # link to other projects’ documentation
    "sphinx.ext.mathjax",  # math support for HTML outputs in Sphinx
    "sphinx_gallery.gen_gallery",  # HTML gallery of examples
    "sphinx_multiversion",  # self-hosted versioned documentation
    "numpydoc",  # support for NumPy style docstrings
    "gh_substitutions",  # custom GitHub substitutions
]

# Add any paths that contain templates here, relative to this directory.
templates_path = [
    "_templates",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

nitpicky = True

nitpick_ignore_regex = [
    ("py:class", "numpy._typing.*"),
    ("py:obj", "trimboth"),  # python 3.11 error
    ("py:obj", "tmean"),  # python 3.11 error
    ("py:obj", "subclass"),  # python 3.11 error
    ("py:class", "typing.Any"),  # python 3.11 error
    # ('py:class', 'numpy.typing.ArrayLike')
    ("py:obj", "sqlalchemy.engine.Engine"),  # ignore sqlalchemy
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

html_logo = "./images/junifer_logo.png"

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "css/custom.css",
]

html_js_files = [
    "js/custom.js",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "versions.html",
        "sidebar/ethical-ads.html",
        "sidebar/scroll-end.html",
    ]
}

# -- sphinx.ext.autodoc configuration ----------------------------------------

autoclass_content = "both"
autodoc_typehints = "description"
autodoc_typehints_description_target = "documented"

# -- sphinx.ext.autosummary configuration ------------------------------------

autosummary_generate = True

# -- sphinx.ext.intersphinx configuration ------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "nilearn": ("https://nilearn.github.io/stable/", None),
    "julearn": ("https://juaml.github.io/julearn/main", None),
    "nibabel": ("https://nipy.org/nibabel/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/dev", None),
    # "sqlalchemy": ("https://docs.sqlalchemy.org/en/20/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

# -- numpydoc configuration --------------------------------------------------

numpydoc_show_class_members = False
numpydoc_xref_param_type = True
numpydoc_xref_aliases = {
    "Path": "pathlib.Path",
    "Nifti1Image": "nibabel.nifti1.Nifti1Image",
    "Nifti2Image": "nibabel.nifti2.Nifti2Image",
    # "Engine": "sqlalchemy.engine.Engine",
}
numpydoc_xref_ignore = {
    "of",
    "shape",
    "optional",
    "or",
    "the",
    "options",
    "function",
    "object",
    "class",
    "objects",
    "Engine",
}
# numpydoc_validation_checks = {
#     "all",
#     "GL01",
#     "GL02",
#     "GL03",
#     "ES01",
#     "SA01",
#     "EX01",
# }

# -- Sphinx-Gallery configuration --------------------------------------------

sphinx_gallery_conf = {
    "examples_dirs": ["../examples"],
    "gallery_dirs": ["auto_examples"],
    "filename_pattern": "/(plot|run)_",
    "backreferences_dir": "generated",
}

# -- sphinx-multiversion configuration ---------------------------------------

smv_rebuild_tags = False
smv_tag_whitelist = r"^v\d+\.\d+.\d+$"
smv_branch_whitelist = r"main"
smv_released_pattern = r"^tags/v.*$"



"""Provide class for IXI VBM juseless datalad datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import List, Union

from ....api.decorators import register_datagrabber
from ....datagrabber import PatternDataladDataGrabber
from ....utils import raise_error


@register_datagrabber
class JuselessDataladIXIVBM(PatternDataladDataGrabber):
    """Juseless IXI VBM Data Grabber class.

    Implements a Data Grabber to access the IXI VBM data in Juseless.

    Parameters
    ----------
    datadir : str or pathlib.Path, optional
        The directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).
    sites : {"Guys", "HH", "IOP"} or list of the options, optional.
        Which sites to access data from. If None, all available sites are
        selected (default None).
    """

    def __init__(
        self,
        datadir: Union[str, Path, None] = None,
        sites: Union[str, List[str], None] = None,
    ) -> None:
        uri = (
            "ria+http://cat_12.5.ds.inm7.de"
            "#b7107c52-8408-11ea-89c6-a0369f287950"
        )
        types = ["VBM_GM"]
        replacements = ["site", "subject"]
        patterns = {
            "VBM_GM": "{site}/sub-{subject}/mri/m0wp1sub-{subject}.nii.gz"
        }

        # validate and/or transform 'site' input
        all_sites = ["HH", "Guys", "IOP"]
        if sites is None:
            sites = all_sites

        if isinstance(sites, str):
            sites = [sites]

        for s in sites:
            if s not in all_sites:
                raise_error(
                    f"{s} not a valid site in IXI VBM dataset!"
                    f"Available sites are {all_sites}"
                )
        self.sites = sites
        super().__init__(
            types=types,
            datadir=datadir,
            uri=uri,
            replacements=replacements,
            patterns=patterns,
        )



"""Provide class for functional connectivity."""

# Authors: Amir Omidvarnia <a.omidvarnia@fz-juelich.de>
#          Kaustubh R. Patil <k.patil@fz-juelich.de>
# License: AGPL

from typing import Any, Dict, List, Optional, Union

from nilearn.connectome import ConnectivityMeasure
from sklearn.covariance import EmpiricalCovariance

from ..api.decorators import register_marker
from .base import BaseMarker
from .parcel_aggregation import ParcelAggregation


@register_marker
class FunctionalConnectivityParcels(BaseMarker):
    """Class for functional connectivity.

    Parameters
    ----------
    parcellation : str or list of str
        The name(s) of the parcellation(s). Check valid options by calling
        :func:`junifer.data.parcellations.list_parcellations`.
    agg_method : str, optional
        The method to perform aggregation using. Check valid options in
        :func:`junifer.stats.get_aggfunc_by_name` (default "mean").
    agg_method_params : dict, optional
        Parameters to pass to the aggregation function. Check valid options in
        :func:`junifer.stats.get_aggfunc_by_name` (default None).
    cor_method : str, optional
        The method to perform correlation using. Check valid options in
        :class:`nilearn.connectome.ConnectivityMeasure`
        (default "covariance").
    cor_method_params : dict, optional
        Parameters to pass to the correlation function. Check valid options in
        :class:`nilearn.connectome.ConnectivityMeasure` (default None).
    mask : str, optional
        The name of the mask to apply to regions before extracting signals.
        Check valid options by calling :func:`junifer.data.masks.list_masks`
        (default None).
    name : str, optional
        The name of the marker. If None, will use the class name (default
        None).
    """

    _DEPENDENCIES = {"nilearn", "scikit-learn"}

    def __init__(
        self,
        parcellation: Union[str, List[str]],
        agg_method: str = "mean",
        agg_method_params: Optional[Dict] = None,
        cor_method: str = "covariance",
        cor_method_params: Optional[Dict] = None,
        mask: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        self.parcellation = parcellation
        self.agg_method = agg_method
        self.agg_method_params = agg_method_params
        self.cor_method = cor_method
        self.cor_method_params = cor_method_params or {}

        # default to nilearn behavior
        self.cor_method_params["empirical"] = self.cor_method_params.get(
            "empirical", False
        )
        self.mask = mask

        super().__init__(name=name)

    def get_valid_inputs(self) -> List[str]:
        """Get valid data types for input.

        Returns
        -------
        list of str
            The list of data types that can be used as input for this marker.
        """
        return ["BOLD"]

    def get_output_type(self, input_type: str) -> str:
        """Get output type.

        Parameters
        ----------
        input_type : str
            The data type input to the marker.

        Returns
        -------
        str
            The storage type output by the marker.

        """
        return "matrix"

    def compute(
        self,
        input: Dict[str, Any],
        extra_input: Optional[Dict] = None,
    ) -> Dict:
        """Compute.

        Parameters
        ----------
        input : dict
            A single input from the pipeline data object in which to compute
            the marker.
        extra_input : dict, optional
            The other fields in the pipeline data object. Useful for accessing
            other data kind that needs to be used in the computation. For
            example, the functional connectivity markers can make use of the
            confounds if available (default None).

        Returns
        -------
        dict
            The computed result as dictionary. The following data will be
            included in the dictionary:

            * ``data`` : functional connectivity matrix as a numpy.ndarray.
            * ``row_names`` : row names as a list
            * ``col_names`` : column names as a list
            * ``matrix_kind`` : the kind of matrix (tril, triu or full)

        """
        pa = ParcelAggregation(
            parcellation=self.parcellation,
            method=self.agg_method,
            method_params=self.agg_method_params,
            mask=self.mask,
            on="BOLD",
        )
        # get the 2D timeseries after parcel aggregation
        ts = pa.compute(input)

        if self.cor_method_params["empirical"]:
            cm = ConnectivityMeasure(
                cov_estimator=EmpiricalCovariance(),  # type: ignore
                kind=self.cor_method,
            )
        else:
            cm = ConnectivityMeasure(kind=self.cor_method)
        out = {}
        out["data"] = cm.fit_transform([ts["data"]])[0]
        # create column names
        out["row_names"] = ts["columns"]
        out["col_names"] = ts["columns"]
        out["matrix_kind"] = "tril"
        return out



"""Provide tests for parser."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import sys
from pathlib import Path

import pytest

from junifer.api.parser import parse_yaml


def test_parse_yaml_failure() -> None:
    """Test YAML parsing failure."""
    with pytest.raises(ValueError, match="does not exist"):
        parse_yaml("foo.yaml")


def test_parse_yaml_success(tmp_path: Path) -> None:
    """Test YAML parsing success.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    # Write test file
    fname = tmp_path / "test_parse_yaml_success.yaml"
    fname.write_text("foo: bar")
    # Check test file
    contents = parse_yaml(fname)
    assert "foo" in contents
    assert contents["foo"] == "bar"


def test_parse_yaml_success_with_module_autoload(tmp_path: Path) -> None:
    """Test YAML parsing with single module autoload success.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    # Write test file
    fname = tmp_path / "test_parse_yaml_with_single_module_autoload.yaml"
    fname.write_text("foo: bar\nwith: numpy")
    # Check test file
    contents = parse_yaml(fname)
    assert "foo" in contents
    assert contents["foo"] == "bar"
    assert "with" in contents
    assert contents["with"] == "numpy"
    assert "numpy" in sys.modules
    assert "junifer.configs.wrong_config" not in sys.modules


def test_parse_yaml_failure_with_multi_module_autoload(tmp_path: Path) -> None:
    """Test YAML parsing with multi module autoload failure.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    # Write test file
    fname = tmp_path / "test_parse_yaml_with_multi_module_autoload.yaml"
    fname.write_text(
        "foo: bar\nwith:\n  - numpy\n  - junifer.testing.wrong_config"
    )
    # Check test file
    with pytest.raises(ImportError, match="wrong_config"):
        parse_yaml(fname)



"""Provide tests for sqlite."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal
from sqlalchemy import create_engine

from junifer.storage.sqlite import SQLiteFeatureStorage
from junifer.storage.utils import element_to_prefix, process_meta


df1 = pd.DataFrame(
    {
        "subject": [1, 2, 3, 4, 5],
        "pk2": ["a", "b", "c", "d", "e"],
        "col1": [11, 22, 33, 44, 55],
        "col2": [111, 222, 333, 444, 555],
    }
).set_index(["subject", "pk2"])

df2 = pd.DataFrame(
    {
        "subject": [2, 5, 6],
        "pk2": ["b", "e", "f"],
        "col1": [2222, 5555, 66],
        "col2": [22222, 55555, 666],
    }
).set_index(["subject", "pk2"])

df_update = pd.DataFrame(
    {
        "subject": [1, 2, 3, 4, 5, 6],
        "pk2": ["a", "b", "c", "d", "e", "f"],
        "col1": [11, 2222, 33, 44, 5555, 66],
        "col2": [111, 22222, 333, 444, 55555, 666],
    }
).set_index(["subject", "pk2"])

df_ignore = pd.DataFrame(
    {
        "subject": [1, 2, 3, 4, 5, 6],
        "pk2": ["a", "b", "c", "d", "e", "f"],
        "col1": [11, 22, 33, 44, 55, 66],
        "col2": [111, 222, 333, 444, 555, 666],
    }
).set_index(["subject", "pk2"])


def _read_sql(
    table_name: str, uri: str, index_col: Union[str, List[str]]
) -> pd.DataFrame:
    """Read database table into a pandas DataFrame.

    Parameters
    ----------
    table_name : str
         The table name.
    uri : str
         The URI of the database.
    index_col : str
         The index column name.

    Returns
    -------
    pandas.DataFrame
        The contents of the table in a DataFrame.

    """
    engine = create_engine(f"sqlite:///{uri}", echo=False)
    df = pd.read_sql(sql=table_name, con=engine, index_col=index_col)
    return df


def test_get_engine_single_output(tmp_path: Path) -> None:
    """Test engine retrieval with single output.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_single_output.sqlite"
    # Single storage, must be the uri
    storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    assert storage.single_output is True
    engine = storage.get_engine()
    assert engine.url.drivername == "sqlite"
    assert engine.url.database == str(uri.absolute())


def test_get_engine_multi_output(tmp_path: Path) -> None:
    """Test engine retrieval with multi output.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_multi_output.sqlite"
    storage = SQLiteFeatureStorage(
        uri=uri, single_output=False, upsert="ignore"
    )
    with pytest.raises(ValueError, match="element must be specified"):
        storage.get_engine()


def test_get_engine_single_output_creation(tmp_path: Path) -> None:
    """Test engine retrieval with single output creation.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    tocreate = tmp_path / "tocreate"
    # Path does not exist yet
    assert not tocreate.exists()
    uri = tocreate.absolute() / "test_single_output.sqlite"
    _ = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    # Path exists now
    assert tocreate.exists()


def test_upsert_replace(tmp_path: Path) -> None:
    """Test dataframe store with upsert=replace.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_upsert_replace.sqlite"
    # Single storage, must be the uri
    storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    # Save to database
    storage.store_df(
        meta_md5="table_name", element={"subject": "test"}, df=df1
    )
    table_name = "meta_table_name"
    # Read stored table
    c_df1 = _read_sql(
        table_name=table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(df1, c_df1)
    # Upsert using replace
    storage._save_upsert(df=df2, name=table_name, if_exists="replace")
    # Read stored table
    c_df2 = _read_sql(
        table_name=table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(df2, c_df2)


def test_upsert_ignore(tmp_path: Path) -> None:
    """Test dataframe store with upsert=ignore.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_upsert_ignore.sqlite"
    # Single storage, must be the uri
    storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    # Save to database
    storage.store_df(
        meta_md5="table_name", element={"subject": "test"}, df=df1
    )
    table_name = "meta_table_name"
    # Read stored table
    c_df1 = _read_sql(
        table_name=table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(df1, c_df1)
    # Check for warning
    with pytest.warns(RuntimeWarning, match="are already present"):
        storage.store_df(
            meta_md5="table_name", element={"subject": "test"}, df=df2
        )
    # Read stored table
    c_dfignore = _read_sql(
        table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(c_dfignore, df_ignore)
    # Check for error
    with pytest.raises(ValueError, match=r"already exists"):
        storage._save_upsert(df2, table_name, if_exists="fail")


def test_upsert_update(tmp_path: Path) -> None:
    """Test dataframe store with upsert=delete.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_upsert_delete.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Save to database
    storage.store_df(
        meta_md5="table_name", element={"subject": "test"}, df=df1
    )
    table_name = "meta_table_name"
    # Read stored table
    c_df1 = _read_sql(
        table_name=table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(df1, c_df1)
    # Save to database
    storage.store_df(
        meta_md5="table_name", element={"subject": "test"}, df=df2
    )
    # Read stored table
    c_dfupdate = _read_sql(
        table_name, uri=uri.as_posix(), index_col=["subject", "pk2"]
    )
    # Check if dataframes are equal
    assert_frame_equal(c_dfupdate, df_update)


def test_upsert_invalid_option(tmp_path: Path) -> None:
    """Test dataframe store with invalid option for upsert.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_upsert_invalid.sqlite"
    with pytest.raises(ValueError):
        SQLiteFeatureStorage(uri=uri, upsert="wrong")


# TODO: can the tests be separated?
def test_store_df_and_read_df(tmp_path: Path) -> None:
    """Test storing dataframe and reading of stored table into dataframe.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_store_df_and_read_df.sqlite"
    storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    # Metadata to store
    meta_md5 = "feature_md5"
    element = {"subject": "test"}
    meta = {
        "element": element,
        "dependencies": ["numpy"],
        "marker": {
            "name": "markername",
        },
        "type": "BOLD",
    }

    _, meta_to_store, element_to_store = process_meta(meta)
    # Store metadata
    storage.store_metadata(
        meta_md5="feature_md5", element=element_to_store, meta=meta_to_store
    )

    # Columns to store
    to_store = df1[["col1", "col2"]]
    # Check for error while storing
    with pytest.raises(ValueError, match=r"missing index items"):
        storage.store_df(
            meta_md5=meta_md5, element=element, df=to_store.set_index("col1")
        )
    # Set index
    to_store = df1.reset_index().set_index(["subject", "pk2", "col1"])
    # Check for error while storing
    with pytest.raises(ValueError, match=r"extra items"):
        storage.store_df(meta_md5=meta_md5, element=element, df=to_store)
    # Convert element to index
    idx = storage.element_to_index(element=element, n_rows=len(to_store))
    # Set index
    to_store = to_store.set_index(idx)
    # Store dataframe
    storage.store_df(meta_md5=meta_md5, element=element, df=to_store)

    # List stored features
    features = storage.list_features()
    # Check correct usage
    assert len(features) == 1
    assert "feature_md5" in features
    # Check for missing feature
    with pytest.raises(ValueError, match="not found"):
        storage.read_df(feature_md5="wrong_md5")
    with pytest.raises(ValueError, match="not found"):
        storage.read_df(feature_name="wrong_name")
    # Check for missing feature to fetch
    with pytest.raises(ValueError, match="least one"):
        storage.read_df()
    # Check for multiple features to fetch
    with pytest.raises(ValueError, match="Only one"):
        storage.read_df(feature_name="wrong_name", feature_md5="wrong_md5")
    # Get MD5 hash of features
    feature_md5 = list(features.keys())[0]
    assert "feature_md5" == feature_md5
    # Check for key
    assert "BOLD_markername" == features[feature_md5]["name"]
    # Read into dataframes
    read_df1 = storage.read_df(feature_md5=feature_md5)
    read_df2 = storage.read_df(feature_name="BOLD_markername")
    # Check if dataframes are equal
    assert_frame_equal(read_df1, read_df2)
    assert_frame_equal(read_df1, to_store)


def test_store_metadata(tmp_path: Path) -> None:
    """Test metadata store.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_metadata_store.sqlite"
    # Single storage, must be the uri
    storage = SQLiteFeatureStorage(uri=uri, upsert="ignore")
    # Metadata to store
    meta = {
        "element": {"subject": "test"},
        "dependencies": ["numpy"],
        "marker": {"name": "test"},
        "type": "BOLD",
    }

    meta_md5, meta_to_store, element_to_store = process_meta(meta)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert meta_md5 == feature_md5


def test_store_table(tmp_path: Path) -> None:
    """Test table store.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_store_table.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Metadata to store
    element = {"subject": "test"}
    dependencies = ["numpy"]
    meta = {
        "element": element,
        "dependencies": dependencies,
        "marker": {"name": "fc"},
        "type": "BOLD",
    }

    meta_md5, meta_to_store, element_to_store = process_meta(meta)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )

    # Data to store
    data = [
        [1, 10],
        [2, 20],
        [3, 30],
        [4, 40],
        [5, 50],
    ]
    # Convert element to index
    idx = storage.element_to_index(element, n_rows=5, rows_col_name="scan")
    # Create dataframe
    df = pd.DataFrame(data, columns=["f1", "f2"], index=idx)

    # Store table
    storage.store_table(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    # Read stored table
    c_df = _read_sql(
        table_name=f"meta_{meta_md5}",
        uri=uri.as_posix(),
        index_col=["subject", "scan"],
    )
    # Check if dataframes are equal
    assert_frame_equal(df, c_df)

    # New data to store
    data_new = [[1, 10], [2, 20], [3, 300], [4, 40], [5, 50], [6, 600]]
    # Convert element to index
    idx_new = storage.element_to_index(element, n_rows=6, rows_col_name="scan")
    # Create dataframe
    df_new = pd.DataFrame(data_new, columns=["f1", "f2"], index=idx_new)
    # Check warning
    with pytest.warns(RuntimeWarning, match=r"Some rows"):
        # Store table
        storage.store_table(
            meta_md5=meta_md5,
            element=element_to_store,
            data=data_new,
            columns=["f1", "f2"],
            rows_col_name="scan",
        )
    # Read stored table
    c_df_new = _read_sql(
        table_name=f"meta_{meta_md5}",
        uri=uri.as_posix(),
        index_col=["subject", "scan"],
    )
    # Check if dataframes are equal
    assert_frame_equal(df_new, c_df_new)


def test_store_matrix(tmp_path: Path) -> None:
    """Test matrix store.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_store_table.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Metadata to store
    element = {"subject": "test"}
    dependencies = ["numpy"]
    meta = {
        "element": element,
        "dependencies": dependencies,
        "marker": {"name": "fc"},
        "type": "BOLD",
    }

    meta_md5, meta_to_store, element_to_store = process_meta(meta)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )

    # Store 4 x 3 full matrix
    data = np.array(
        [[1, 2, 3], [11, 22, 33], [111, 222, 333], [1111, 2222, 3333]]
    )
    row_names = ["row1", "row2", "row3", "row4"]
    col_names = ["col1", "col2", "col3"]

    # Store table
    storage.store_matrix(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        row_names=row_names,
        col_names=col_names,
    )
    stored_names = [f"{i}~{j}" for i in row_names for j in col_names]

    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]

    read_df = storage.read_df(feature_md5=feature_md5)
    assert read_df.shape == (1, 12)
    assert_array_equal(read_df.values[0], data.flatten())
    assert list(read_df.columns) == stored_names
    # Store without row and column names
    uri = tmp_path / "test_store_table_nonames.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    # Store table
    storage.store_matrix(
        meta_md5=meta_md5, element=element_to_store, data=data
    )
    stored_names = [
        f"r{i}~c{j}"
        for i in range(data.shape[0])
        for j in range(data.shape[1])
    ]
    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]
    read_df = storage.read_df(feature_md5=feature_md5)
    assert list(read_df.columns) == stored_names

    with pytest.raises(ValueError, match="Invalid kind"):
        storage.store_matrix(
            meta_md5=meta_md5,
            element=element_to_store,
            data=data,
            row_names=row_names,
            col_names=col_names,
            matrix_kind="wrong",
        )

    with pytest.raises(ValueError, match="non-square"):
        storage.store_matrix(
            meta_md5=meta_md5,
            element=element_to_store,
            data=data,
            row_names=row_names,
            col_names=col_names,
            matrix_kind="triu",
        )

    with pytest.raises(ValueError, match="cannot be False"):
        storage.store_matrix(
            meta_md5=meta_md5,
            element=element_to_store,
            data=data,
            row_names=row_names,
            col_names=col_names,
            matrix_kind="full",
            diagonal=False,
        )

    # Store upper triangular matrix
    data = np.array([[1, 2, 3], [11, 22, 33], [111, 222, 333]])
    row_names = ["row1", "row2", "row3"]
    col_names = ["col1", "col2", "col3"]
    uri = tmp_path / "test_store_table_triu.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    storage.store_matrix(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        row_names=row_names,
        col_names=col_names,
        matrix_kind="triu",
    )

    stored_names = [
        "row1~col1",
        "row1~col2",
        "row1~col3",
        "row2~col2",
        "row2~col3",
        "row3~col3",
    ]

    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]
    read_df = storage.read_df(feature_md5=feature_md5)
    assert list(read_df.columns) == stored_names
    assert_array_equal(
        read_df.values, data[np.triu_indices(n=data.shape[0])][None, :]
    )

    # Store upper triangular matrix without diagonal
    uri = tmp_path / "test_store_table_triu_nodiagonal.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    storage.store_matrix(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        row_names=row_names,
        col_names=col_names,
        matrix_kind="triu",
        diagonal=False,
    )

    stored_names = [
        "row1~col2",
        "row1~col3",
        "row2~col3",
    ]

    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]
    read_df = storage.read_df(feature_md5=feature_md5)
    assert list(read_df.columns) == stored_names
    assert_array_equal(
        read_df.values, data[np.triu_indices(n=data.shape[0], k=1)][None, :]
    )

    # Store lower triangular matrix
    data = np.array([[1, 2, 3], [11, 22, 33], [111, 222, 333]])
    row_names = ["row1", "row2", "row3"]
    col_names = ["col1", "col2", "col3"]
    uri = tmp_path / "test_store_table_tril.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    storage.store_matrix(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        row_names=row_names,
        col_names=col_names,
        matrix_kind="tril",
    )

    stored_names = [
        "row1~col1",
        "row2~col1",
        "row2~col2",
        "row3~col1",
        "row3~col2",
        "row3~col3",
    ]

    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]
    read_df = storage.read_df(feature_md5=feature_md5)
    assert list(read_df.columns) == stored_names
    assert_array_equal(
        read_df.values, data[np.tril_indices(n=data.shape[0])][None, :]
    )

    # Store lower triangular matrix without diagonal
    uri = tmp_path / "test_store_table_tril_nodiagonal.sqlite"
    storage = SQLiteFeatureStorage(uri=uri)
    # Store metadata
    storage.store_metadata(
        meta_md5=meta_md5, element=element_to_store, meta=meta_to_store
    )
    storage.store_matrix(
        meta_md5=meta_md5,
        element=element_to_store,
        data=data,
        row_names=row_names,
        col_names=col_names,
        matrix_kind="tril",
        diagonal=False,
    )
    stored_names = [
        "row2~col1",
        "row3~col1",
        "row3~col2",
    ]

    features = storage.list_features()
    feature_md5 = list(features.keys())[0]
    assert "BOLD_fc" == features[feature_md5]["name"]
    read_df = storage.read_df(feature_md5=feature_md5)
    assert list(read_df.columns) == stored_names
    assert_array_equal(
        read_df.values, data[np.tril_indices(n=data.shape[0], k=-1)][None, :]
    )


# TODO: can the test be parametrized?
def test_store_multiple_output(tmp_path: Path):
    """Test storing using single_output=False.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_store_multiple_output.sqlite"
    storage = SQLiteFeatureStorage(uri=uri, single_output=False)
    # Metadata to store
    meta1 = {
        "element": {"subject": "test-01", "session": "ses-01"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    meta2 = {
        "element": {"subject": "test-02", "session": "ses-01"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    meta3 = {
        "element": {"subject": "test-01", "session": "ses-02"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    # Data to store
    data1 = np.array(
        [
            [1, 10],
            [2, 20],
            [3, 30],
            [4, 40],
            [5, 50],
        ]
    )
    data2 = data1 * 10
    data3 = data1 * 20
    # Process metadata for storage
    hash1, meta_to_store1, element_to_store1 = process_meta(meta1)
    # Convert element to index
    idx1 = storage.element_to_index(
        element_to_store1, n_rows=5, rows_col_name="scan"
    )
    # Create dataframe
    df1 = pd.DataFrame(data1, columns=["f1", "f2"], index=idx1)
    # Process metadata for storage
    hash2, meta_to_store2, element_to_store2 = process_meta(meta2)
    # Convert element to index
    idx2 = storage.element_to_index(
        element_to_store2, n_rows=5, rows_col_name="scan"
    )
    # Create dataframe
    df2 = pd.DataFrame(data2, columns=["f1", "f2"], index=idx2)
    # Process metadata for storage
    hash3, meta_to_store3, element_to_store3 = process_meta(meta3)
    # Convert element to index
    idx3 = storage.element_to_index(
        element_to_store3, n_rows=5, rows_col_name="scan"
    )
    # Create dataframe
    df3 = pd.DataFrame(data3, columns=["f1", "f2"], index=idx3)
    # Check hash equality
    assert hash1 == hash2
    assert hash2 == hash3
    # Store tables
    storage.store_metadata(
        meta_md5=hash1, element=element_to_store1, meta=meta_to_store1
    )
    storage.store_metadata(
        meta_md5=hash2, element=element_to_store2, meta=meta_to_store2
    )
    storage.store_metadata(
        meta_md5=hash3, element=element_to_store3, meta=meta_to_store3
    )
    storage.store_table(
        meta_md5=hash1,
        element=element_to_store1,
        data=data1,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    storage.store_table(
        meta_md5=hash2,
        element=element_to_store2,
        data=data2,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    storage.store_table(
        meta_md5=hash3,
        element=element_to_store3,
        data=data3,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    # Check that URI does not exist yet
    assert not uri.exists()
    # Convert element to preifx
    prefix1 = element_to_prefix(meta1["element"])
    prefix2 = element_to_prefix(meta2["element"])
    prefix3 = element_to_prefix(meta3["element"])
    # URIs for data storage
    uri1 = uri.parent / f"{prefix1}{uri.name}"
    uri2 = uri.parent / f"{prefix2}{uri.name}"
    uri3 = uri.parent / f"{prefix3}{uri.name}"
    # Check URIs for data storage exist
    assert uri1.exists()
    assert uri2.exists()
    assert uri3.exists()
    # Set index columns
    cols = ["subject", "session", "scan"]
    table_name = f"meta_{hash1}"
    # Read stored tables
    cdf1 = _read_sql(table_name, uri1.as_posix(), index_col=cols)
    cdf2 = _read_sql(table_name, uri2.as_posix(), index_col=cols)
    cdf3 = _read_sql(table_name, uri3.as_posix(), index_col=cols)
    # Check if dataframes are equal
    assert_frame_equal(df1, cdf1)
    assert_frame_equal(df2, cdf2)
    assert_frame_equal(df3, cdf3)


# TODO: can test be paramtrized?
def test_collect(tmp_path: Path) -> None:
    """Test collect.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    uri = tmp_path / "test_collect.sqlite"
    storage = SQLiteFeatureStorage(uri=uri, single_output=False)
    # Metadata for storage
    meta1 = {
        "element": {"subject": "test-01", "session": "ses-01"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    meta2 = {
        "element": {"subject": "test-02", "session": "ses-01"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    meta3 = {
        "element": {"subject": "test-01", "session": "ses-02"},
        "dependencies": ["numpy"],
        "marker": {"name": "fc"},
        "type": "BOLD",
    }
    # Data for storage
    data1 = np.array(
        [
            [1, 10],
            [2, 20],
            [3, 30],
            [4, 40],
            [5, 50],
        ]
    )
    data2 = data1 * 10
    data3 = data1 * 20
    # Store tables
    hash1, meta_to_store1, element_to_store1 = process_meta(meta1)
    hash2, meta_to_store2, element_to_store2 = process_meta(meta2)
    hash3, meta_to_store3, element_to_store3 = process_meta(meta3)
    storage.store_metadata(
        meta_md5=hash1, element=element_to_store1, meta=meta_to_store1
    )
    storage.store_metadata(
        meta_md5=hash2, element=element_to_store2, meta=meta_to_store2
    )
    storage.store_metadata(
        meta_md5=hash3, element=element_to_store3, meta=meta_to_store3
    )
    storage.store_table(
        meta_md5=hash1,
        element=element_to_store1,
        data=data1,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    storage.store_table(
        meta_md5=hash2,
        element=element_to_store2,
        data=data2,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    storage.store_table(
        meta_md5=hash3,
        element=element_to_store3,
        data=data3,
        columns=["f1", "f2"],
        rows_col_name="scan",
    )
    # Convert element to prefix
    prefix1 = element_to_prefix(meta1["element"])
    prefix2 = element_to_prefix(meta2["element"])
    prefix3 = element_to_prefix(meta3["element"])
    # URIs for data storage
    uri1 = uri.parent / f"{prefix1}{uri.name}"
    uri2 = uri.parent / f"{prefix2}{uri.name}"
    uri3 = uri.parent / f"{prefix3}{uri.name}"
    # Check URIs for data storage exist
    assert uri1.exists()
    assert uri2.exists()
    assert uri3.exists()
    # Check that URI does not exist yet
    assert not uri.exists()
    # Collect data
    storage.collect()
    # Check that URI exists now
    assert uri.exists()
    # Set index columns
    cols = ["subject", "session", "scan"]
    # Store metadata
    table_name = f"meta_{hash1}"
    # Read stored tables
    all_df = _read_sql(table_name, uri.as_posix(), index_col=cols)
    cdf1 = _read_sql(table_name, uri1.as_posix(), index_col=cols)
    cdf2 = _read_sql(table_name, uri2.as_posix(), index_col=cols)
    cdf3 = _read_sql(table_name, uri3.as_posix(), index_col=cols)
    # Operate on retrieved tables
    all_cdf = pd.concat([cdf1, cdf2, cdf3])
    all_df.sort_index(level=cols, inplace=True)
    all_cdf.sort_index(level=cols, inplace=True)
    # Check if dataframes are equal
    assert_frame_equal(all_df, all_cdf)



"""Provide imports for api sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from . import decorators
from .cli import cli
from .functions import collect, queue, run



"""Provide testing datagrabbers."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import tempfile
from pathlib import Path
from typing import Dict, List

import nibabel as nib
from nilearn import datasets, image

from ..datagrabber.base import BaseDataGrabber


class OasisVBMTestingDatagrabber(BaseDataGrabber):
    """Data Grabber for Oasis VBM testing data."""

    def __init__(self) -> None:
        # Create temporary directory
        datadir = tempfile.mkdtemp()
        # Define types
        types = ["VBM_GM"]
        super().__init__(types=types, datadir=datadir)

    def get_element_keys(self) -> List[str]:
        """Get element keys.

        Returns
        -------
        list of str
            The element keys.

        """
        return ["subject"]

    def get_item(self, subject: str) -> Dict[str, Dict]:
        """Implement indexing support.

        Parameters
        ----------
        subject : str
            The subject to retrieve.

        Returns
        -------
        dict
            The data along with the metadata.

        """
        out = {}
        i_sub = int(subject.split("-")[1]) - 1
        out["VBM_GM"] = {"path": Path(self._dataset.gray_matter_maps[i_sub])}

        return out

    def __enter__(self) -> "OasisVBMTestingDatagrabber":
        """Implement context entry.

        Returns
        -------
        OasisVBMTestingDatagrabber

        """
        self._dataset = datasets.fetch_oasis_vbm(n_subjects=10)
        return self

    def get_elements(self) -> List[str]:
        """Get elements.

        Returns
        -------
        list of str
            List of elements that can be grabbed.

        """
        return [f"sub-{x:02d}" for x in list(range(1, 11))]


class SPMAuditoryTestingDatagrabber(BaseDataGrabber):
    """Data Grabber for SPM Auditory dataset.

    Wrapper for :func:`nilearn.datasets.fetch_spm_auditory`.

    """

    def __init__(self) -> None:
        # Create temporary directory
        datadir = tempfile.mkdtemp()
        # Define types
        types = ["BOLD", "T1w"]  # TODO: Check that they are T1w
        super().__init__(types=types, datadir=datadir)

    def get_element_keys(self) -> List[str]:
        """Get element keys.

        Returns
        -------
        list of str
            The element keys.

        """
        return ["subject"]

    def get_elements(self) -> List[str]:
        """Get elements.

        Returns
        -------
        list of str
            List of elements that can be grabbed.

        """
        return [f"sub{x:03d}" for x in list(range(1, 11))]

    def get_item(self, subject: str) -> Dict[str, Dict]:
        """Implement indexing support.

        Parameters
        ----------
        subject : str
            The subject to retrieve.

        Returns
        -------
        dict
            The data along with the metadata.

        """
        out = {}
        nilearn_data = datasets.fetch_spm_auditory(subject_id=subject)
        fmri_img = image.concat_imgs(nilearn_data.func)  # type: ignore
        anat_img = image.concat_imgs(nilearn_data.anat)  # type: ignore

        fmri_fname = self.datadir / f"{subject}_bold.nii.gz"
        anat_fname = self.datadir / f"{subject}_T1w.nii.gz"
        nib.save(fmri_img, fmri_fname)
        nib.save(anat_img, anat_fname)
        out["BOLD"] = {"path": fmri_fname}
        out["T1w"] = {"path": anat_fname}
        return out


class PartlyCloudyTestingDataGrabber(BaseDataGrabber):
    """Data Grabber for Partly Cloudy dataset.

    Wrapper for :func:`nilearn.datasets.fetch_development_fmri`

    Parameters
    ----------
    reduce_confounds : bool, optional
        If True, the returned confounds only include 6 motion parameters,
        mean framewise displacement, signal from white matter, csf, and
        6 anatomical compcor parameters. This selection only serves the
        purpose of having realistic examples. Depending on your research
        question, other confounds might be more appropriate.
        If False, returns all :term:`fMRIPrep` confounds (default True).

    age_group : {"adult", "child", "both"}, optional
       Which age group to fetch:

        * ``adult`` : fetch adults only (n=33, ages 18-39)
        * ``child`` : fetch children only (n=122, ages 3-12)
        * ``both`` : fetch full sample (n=155)

        (default "both")

    """

    def __init__(
        self, reduce_confounds: bool = True, age_group: str = "both"
    ) -> None:
        """Initialize the class."""
        datadir = tempfile.mkdtemp()
        # Define types
        types = ["BOLD", "BOLD_confounds"]
        self.reduce_confounds = reduce_confounds
        self.age_group = age_group
        super().__init__(types=types, datadir=datadir)

    def __enter__(self) -> "PartlyCloudyTestingDataGrabber":
        """Implement context entry.

        Returns
        -------
        PartlyCloudyTestingDataGrabber

        """
        self._dataset = datasets.fetch_development_fmri(
            n_subjects=10,
            reduce_confounds=self.reduce_confounds,
            age_group=self.age_group,
        )
        return self

    def get_element_keys(self) -> List[str]:
        """Get element keys.

        Returns
        -------
        list of str
            The element keys.

        """
        return ["subject"]

    def get_elements(self) -> List[str]:
        """Get elements.

        Returns
        -------
        list of str
            List of elements that can be grabbed.

        """
        return [f"sub-{x:02d}" for x in list(range(1, 11))]

    def get_item(self, subject: str) -> Dict[str, Dict]:
        """Implement indexing support.

        Parameters
        ----------
        subject : str
            The subject to retrieve.

        Returns
        -------
        dict
            The data along with the metadata.

        """
        out = {}
        i_sub = int(subject.split("-")[1]) - 1
        out["BOLD"] = {"path": Path(self._dataset["func"][i_sub])}
        conf_format = "fmriprep"

        out["BOLD_confounds"] = {
            "path": Path(self._dataset["confounds"][i_sub]),
            "format": conf_format,
        }

        return out



"""Provide tests for base."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import pytest

from junifer.storage.base import BaseFeatureStorage


def test_BaseFeatureStorage_abstractness() -> None:
    """Test BaseFeatureStorage is abstract base class."""
    with pytest.raises(TypeError, match=r"abstract"):
        BaseFeatureStorage(uri="/tm", storage_types=["matrix"])  # type: ignore


def test_BaseFeatureStorage() -> None:
    """Test proper subclassing of BaseFeatureStorage."""
    # Create concrete class
    class MyFeatureStorage(BaseFeatureStorage):
        """Implement concrete class."""

        def __init__(self, uri, single_output=False):
            storage_types = ["matrix", "table", "timeseries"]
            super().__init__(
                uri=uri,
                storage_types=storage_types,
                single_output=single_output,
            )

        def get_valid_inputs(self):
            return ["matrix", "table", "timeseries"]

        def list_features(self):
            super().list_features()

        def read_df(self, feature_name=None, feature_md5=None):
            super().read_df(
                feature_name=feature_name,
                feature_md5=feature_md5,
            )

        def store_metadata(self, meta_md5, meta, element):
            super().store_metadata(meta_md5, meta, element)

        def collect(self):
            return super().collect()

    # Check single_output is False
    st = MyFeatureStorage(uri="/tmp")
    assert st.single_output is False
    # Check single_output is True
    st = MyFeatureStorage(uri="/tmp", single_output=True)
    assert st.single_output is True

    # Check validate with valid argument
    st.validate(input_=["matrix"])
    # Check validate with invalid argument
    with pytest.raises(ValueError):
        st.validate(input_=["duck"])

    with pytest.raises(NotImplementedError):
        st.list_features()

    with pytest.raises(NotImplementedError):
        st.read_df(None)

    element = {"subject": "test"}
    dependencies = ["numpy"]
    meta = {
        "element": element,
        "dependencies": dependencies,
        "marker": {"name": "fc"},
        "type": "BOLD",
    }

    with pytest.raises(NotImplementedError):
        st.store(kind="matrix", meta=meta)

    with pytest.raises(NotImplementedError):
        st.store_metadata("md5", meta=meta, element={})

    with pytest.raises(NotImplementedError):
        st.collect()

    with pytest.raises(NotImplementedError):
        st.store(kind="timeseries", meta=meta)

    with pytest.raises(NotImplementedError):
        st.store(kind="table", meta=meta)

    with pytest.raises(ValueError):
        st.store(kind="lego", meta=meta)

    assert st.uri == "/tmp"



"""Provide imports for datagrabber sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL


# These 4 need to be in this order, otherwise it is a circular import
from .base import BaseDataGrabber
from .datalad_base import DataladDataGrabber
from .pattern import PatternDataGrabber
from .pattern_datalad import PatternDataladDataGrabber

from .aomic import DataladAOMICID1000, DataladAOMICPIOP1, DataladAOMICPIOP2
from .hcp import HCP1200, DataladHCP1200
from .multiple import MultipleDataGrabber



"""Create an example/testing dataset for PIOP2 with mock data."""
# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Vera Komeyer <v.komeyer@fz-juelich.de>
#          Xuan Li <xu.li@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL
from pathlib import Path
from tempfile import TemporaryDirectory

import datalad.api as dl


# repo has to be created on gin manually beforehand if not owner
dst = "git@gin.g-node.org:/juaml/datalad-example-aomicpiop2.git"

with TemporaryDirectory() as tmpdir_name:
    tmpdir = Path(tmpdir_name)
    ds = dl.create(tmpdir)  # type: ignore

    base_dir = tmpdir / "derivatives"
    base_dir.mkdir(exist_ok=True, parents=True)

    for dtype in ["dwipreproc", "fmriprep"]:
        dtype_dir = base_dir / dtype
        dtype_dir.mkdir()

        for i_sub in range(1, 10):
            t_sub = f"sub-{i_sub:04d}"
            sub_dir = dtype_dir / t_sub
            sub_dir.mkdir()

            if dtype == "fmriprep":
                for dname in ["func", "anat"]:
                    (sub_dir / dname).mkdir()

                fnames = [
                    (
                        f"anat/{t_sub}_space-MNI152NLin2009cAsym_desc-preproc"
                        "_T1w.nii.gz"
                    ),
                    (
                        f"anat/{t_sub}_space-MNI152NLin2009cAsym_label-"
                        "CSF_probseg.nii.gz"
                    ),
                    (
                        f"anat/{t_sub}_space-MNI152NLin2009cAsym_label-"
                        "GM_probseg.nii.gz"
                    ),
                    (
                        f"anat/{t_sub}_space-MNI152NLin2009cAsym_label-"
                        "WM_probseg.nii.gz"
                    ),
                ]

                tasks = [
                    "stopsignal_acq-seq",
                    "emomatching_acq-seq",
                    "restingstate_acq-seq",
                    "workingmemory_acq-seq",
                ]
                for t in tasks:
                    fnames.append(
                        (
                            f"func/{t_sub}_task-{t}_space-"
                            "MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
                        )
                    )
                    fnames.append(
                        (
                            f"func/{t_sub}_task-{t}_space-"
                            "MNI152NLin2009cAsym_desc-preproc_bold.json"
                        )
                    )
                    fnames.append(
                        (
                            f"func/{t_sub}_task-{t}_desc-confounds"
                            "_regressors.tsv"
                        )
                    )
                    fnames.append(
                        (
                            f"func/{t_sub}_task-{t}_desc-confounds"
                            "_regressors.json"
                        )
                    )

            elif dtype == "dwipreproc":
                dname = "dwi"
                (sub_dir / dname).mkdir()

                fnames = [
                    (f"{dname}/{t_sub}_desc-preproc_dwi.nii.gz"),
                ]

            for fname in fnames:
                with open(sub_dir / fname, "w") as f:
                    f.write("placeholder")

    ds.save(recursive=True)
    # use this to create the repo automatically, only possible for juaml owner
    # ds.create_sibling_gin(
    # (org_name/repo_basename).as_posix(), name='gin', existing='reconfigure',
    # api=dst_api, access_protocol='ssh')
    ds.siblings("add", name="gin", url=dst)
    ds.push(to="gin", force="all")



"""Provide tests for utils."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import pytest

from junifer.datagrabber.utils import (
    validate_patterns,
    validate_replacements,
    validate_types,
)


def test_validate_types() -> None:
    """Test validation of types."""
    with pytest.raises(TypeError, match="must be a list"):
        validate_types("wrong")  # type: ignore
    with pytest.raises(TypeError, match="must be a list of strings"):
        validate_types([1])  # type: ignore

    validate_types(["T1w", "BOLD"])


def test_validate_replacements() -> None:
    """Test validation of replacements."""
    with pytest.raises(TypeError, match="must be a list"):
        validate_replacements("wrong", "also wrong")  # type: ignore
    with pytest.raises(TypeError, match="must be a dict"):
        validate_replacements(["correct"], "wrong")  # type: ignore

    patterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
        "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }

    with pytest.raises(TypeError, match="must be a list of strings"):
        validate_replacements([1], patterns)  # type: ignore

    with pytest.raises(ValueError, match="is not part of"):
        validate_replacements(["session"], patterns)

    wrong_patterns = {
        "T1w": "{subject}/anat/_T1w.nii.gz",
        "BOLD": "{session}/func/_task-rest_bold.nii.gz",
    }

    with pytest.raises(ValueError, match="At least one pattern"):
        validate_replacements(["subject", "session"], wrong_patterns)

    validate_replacements(["subject"], patterns)


def test_validate_patterns() -> None:
    """Test validation of patterns."""
    types = ["T1w", "BOLD"]
    with pytest.raises(TypeError, match="must be a dict"):
        validate_patterns(types, "wrong")  # type: ignore

    wrongpatterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
    }

    with pytest.raises(ValueError, match="same length"):
        validate_patterns(types, wrongpatterns)  # type: ignore

    wrongpatterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
        "T2": "{subject}/anat/{subject}_T2.nii.gz",
    }

    with pytest.raises(ValueError, match="contain all"):
        validate_patterns(types, wrongpatterns)  # type: ignore

    patterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
        "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }

    wrongpatterns = {
        "T1w": "{subject}/anat/{subject}*.nii",
        "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }

    with pytest.raises(ValueError, match="following a replacement"):
        validate_patterns(types, wrongpatterns)

    validate_patterns(types, patterns)



"""Provide imports for utils sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from .fs import make_executable
from .logging import configure_logging, logger, raise_error, warn_with_log



"""Provide tests for pandas base feature storage."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from junifer.storage.pandas_base import PandasBaseFeatureStorage


def test_element_to_index() -> None:
    """Test element to index."""
    element = {"foo": "bar"}
    index = PandasBaseFeatureStorage.element_to_index(element)
    assert index.names == ["foo", "idx"]
    assert index.levels[0].name == "foo"
    assert index.levels[0].values[0] == "bar"
    assert all(x == "bar" for x in index.levels[0].values)
    assert index.levels[0].values.shape == (1,)
    assert index.levels[1].name == "idx"
    assert all(x == i for i, x in enumerate(index.levels[1].values))
    assert index.levels[1].values.shape == (1,)

    index = PandasBaseFeatureStorage.element_to_index(element, n_rows=10)
    assert index.names == ["foo", "idx"]
    assert index.levels[0].name == "foo"
    assert all(x == "bar" for x in index.levels[0].values)
    assert index.levels[0].values.shape == (1,)

    assert index.levels[1].name == "idx"
    assert all(x == i for i, x in enumerate(index.levels[1].values))
    assert index.levels[1].values.shape == (10,)

    index = PandasBaseFeatureStorage.element_to_index(
        element, n_rows=1, rows_col_name="scan"
    )
    assert index.names == ["foo", "scan"]
    assert index.levels[0].name == "foo"
    assert index.levels[0].values[0] == "bar"
    assert all(x == "bar" for x in index.levels[0].values)
    assert index.levels[0].values.shape == (1,)
    assert index.levels[1].name == "scan"
    assert all(x == i for i, x in enumerate(index.levels[1].values))
    assert index.levels[1].values.shape == (1,)

    index = PandasBaseFeatureStorage.element_to_index(
        element, n_rows=7, rows_col_name="scan"
    )
    assert index.names == ["foo", "scan"]
    assert index.levels[0].name == "foo"
    assert all(x == "bar" for x in index.levels[0].values)
    assert index.levels[0].values.shape == (1,)

    assert index.levels[1].name == "scan"
    assert all(x == i for i, x in enumerate(index.levels[1].values))
    assert index.levels[1].values.shape == (7,)

    element = {"subject": "sub-01", "session": "ses-01"}
    index = PandasBaseFeatureStorage.element_to_index(element, n_rows=10)

    assert index.levels[0].name == "subject"
    assert all(x == "sub-01" for x in index.levels[0].values)
    assert index.levels[0].values.shape == (1,)

    assert index.levels[1].name == "session"
    assert all(x == "ses-01" for x in index.levels[1].values)
    assert index.levels[1].values.shape == (1,)

    assert index.levels[2].name == "idx"
    assert all(x == i for i, x in enumerate(index.levels[2].values))
    assert index.levels[2].values.shape == (10,)



"""
UKB VBM GMD Extraction
======================

Authors: Federico Raimondo

License: BSD 3 clause
"""
import tempfile

import junifer.testing.registry  # noqa: F401
from junifer.api import run


datagrabber = {
    "kind": "OasisVBMTestingDatagrabber",
}

markers = [
    {
        "name": "Schaefer1000x7_TrimMean80",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "trim_mean",
        "method_params": {"proportiontocut": 0.2},
    },
    {
        "name": "Schaefer1000x7_Mean",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "mean",
    },
    {
        "name": "Schaefer1000x7_Std",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "std",
    },
]

storage = {
    "kind": "SQLiteFeatureStorage",
}

with tempfile.TemporaryDirectory() as tmpdir:
    uri = f"{tmpdir}/test.sqlite"
    storage["uri"] = uri
    run(
        workdir="/tmp",
        datagrabber=datagrabber,
        markers=markers,
        storage=storage,
    )



"""
Run junifer and julearn.
========================

This example uses a ParcelAggregation marker to compute the mean of each parcel
using the Schaefer parcellation (100 rois, 7 Yeo networks) for a 3D nifti to
extract some features for machine learning using julearn to predict some other
data.

Authors: Leonard Sasse, Sami Hamdan, Nicolas Nieto, Synchon Mandal

License: BSD 3 clause
"""

import tempfile

import nilearn
import pandas as pd
from julearn import run_cross_validation

import junifer.testing.registry  # noqa
from junifer.api import collect, run
from junifer.storage.sqlite import SQLiteFeatureStorage
from junifer.utils import configure_logging


###############################################################################
# Set the logging level to info to see extra information:
configure_logging(level="INFO")


###############################################################################
# Define the markers you want:

marker_dicts = [
    {
        "name": "Schaefer100x17_TrimMean80",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer100x17",
        "method": "trim_mean",
        "method_params": {"proportiontocut": 0.2},
    },
    {
        "name": "Schaefer200x17_Mean",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer200x17",
        "method": "mean",
    },
]


###############################################################################
# Define target and confounds for julearn machine learning:
y = "age"
confound = "sex"


###############################################################################
# Load the VBM phenotype data for machine learning data:
# - Fetch the Oasis dataset
oasis_dataset = nilearn.datasets.fetch_oasis_vbm()
age = oasis_dataset.ext_vars[y][:10]
sex = (
    pd.Series(oasis_dataset.ext_vars["mf"][:10])
    .map(lambda x: 1 if x == "F" else 0)
    .values
)


###############################################################################
# Create a temporary directory for junifer feature extraction:
with tempfile.TemporaryDirectory() as tmpdir:

    storage = {"kind": "SQLiteFeatureStorage", "uri": f"{tmpdir}/test.sqlite"}
    # run the defined junifer feature extraction pipeline
    run(
        workdir="/tmp",
        datagrabber={"kind": "OasisVBMTestingDatagrabber"},
        markers=marker_dicts,
        storage=storage,
    )

    # read in extracted features and add confounds and targets
    # for julearn run cross validation
    collect(storage)
    db = SQLiteFeatureStorage(uri=storage["uri"])

    df_vbm = db.read_df(feature_name="VBM_GM_Schaefer200x17_Mean")
    oasis_subjects = [x[0] for x in df_vbm.index]
    df_vbm.index = oasis_subjects


###############################################################################
# Using julearn for machine learning:
# We predict the age given our vbm features and sex as a confound.
X = list(df_vbm.columns)
df_vbm[y] = age
df_vbm[confound] = sex

scores = run_cross_validation(
    X=X,
    confounds=confound,
    y=y,
    data=df_vbm,
    problem_type="regression",
    model="ridge",
    cv=3,
    preprocess_X=["zscore", "remove_confound"],
)
print(scores)

###############################################################################
# Interpretation of results:
# Doing machine learning with only 10 datapoints is not meaningful.
# This explains the big variation in scores
# for different cross-validation folds.



"""Provide decorators for api."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL
from typing import Type

from ..pipeline.registry import register


def register_datagrabber(klass: Type) -> Type:
    """Datagrabber registration decorator.

    Registers the datagrabber so it can be used by name.

    Parameters
    ----------
    klass: class
        The class of the datagrabber to register.

    Returns
    -------
    klass: class
        The unmodified input class.

    """
    register(
        step="datagrabber",
        name=klass.__name__,
        klass=klass,
    )
    return klass


def register_datareader(klass: Type) -> Type:
    """Datareader registration decorator.

    Registers the datareader so it can be used by name.

    Parameters
    ----------
    klass: class
        The class of the datareader to register.

    Returns
    -------
    klass: class
        The unmodified input class.

    """
    register(
        step="datareader",
        name=klass.__name__,
        klass=klass,
    )
    return klass


def register_preprocessor(klass: Type) -> Type:
    """Preprocessor registration decorator.

    Registers the preprocessor so it can be used by name.

    Parameters
    ----------
    klass: class
        The class of the preprocessor to register.

    Returns
    -------
    klass: class
        The unmodified input class.

    """
    register(
        step="preprocessing",
        name=klass.__name__,
        klass=klass,
    )
    return klass


def register_marker(klass: Type) -> Type:
    """Marker registration decorator.

    Registers the marker so it can be used by name.

    Parameters
    ----------
    klass: class
        The class of the marker to register.

    Returns
    -------
    klass: class
        The unmodified input class.

    """
    register(
        step="marker",
        name=klass.__name__,
        klass=klass,
    )
    return klass


def register_storage(klass: Type) -> Type:
    """Storage registration decorator.

    Registers the storage so it can be used by name.

    Parameters
    ----------
    klass: class
        The class of the storage to register.

    Returns
    -------
    klass: class
        The unmodified input class.

    """
    register(
        step="storage",
        name=klass.__name__,
        klass=klass,
    )
    return klass



"""Provide abstract base class for feature storage."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd

from ..utils import raise_error
from .utils import process_meta


class BaseFeatureStorage(ABC):
    """Abstract base class for feature storage.

    For every interface that is required, one needs to provide a concrete
    implementation of this abstract class.

    Parameters
    ----------
    uri : str or pathlib.Path
        The path to the storage.
    storage_types : str or list of str
        The available storage types for the class.
    single_output : bool, optional
        Whether to have single output (default True).

    """

    def __init__(
        self,
        uri: Union[str, Path],
        storage_types: Union[List[str], str],
        single_output: bool = True,
    ) -> None:
        self.uri = uri
        if not isinstance(storage_types, list):
            storage_types = [storage_types]
        if any(x not in self.get_valid_inputs() for x in storage_types):
            wrong_storage_types = [
                x for x in storage_types if x not in self.get_valid_inputs()
            ]
            raise ValueError(
                f"{self.__class__.__name__} cannot store {wrong_storage_types}"
            )
        self._valid_inputs = storage_types
        self.single_output = single_output

    def get_valid_inputs(self) -> List[str]:
        """Get valid storage types for input.

        Returns
        -------
        list of str
            The list of storage types that can be used as input for this "
            "storage.
        """
        raise_error(
            msg="Concrete classes need to implement get_valid_inputs().",
            klass=NotImplementedError,
        )

    def validate(self, input_: List[str]) -> None:
        """Validate the input to the pipeline step.

        Parameters
        ----------
        input_ : list of str
            The input to the pipeline step.

        Raises
        ------
        ValueError
            If the ``input_`` is invalid.

        """
        if not any(x in input_ for x in self._valid_inputs):
            raise_error(
                "Input does not have the required data."
                f"\t Input: {input}"
                f"\t Required (any of): {self._valid_inputs}"
            )

    @abstractmethod
    def list_features(self) -> Dict:
        """List the features in the storage.

        Returns
        -------
        dict
            List of features in the storage. The keys are the feature names to
            be used in read_features() and the values are the metadata of each
            feature.

        """
        raise_error(
            msg="Concrete classes need to implement list_features().",
            klass=NotImplementedError,
        )

    @abstractmethod
    def read_df(
        self,
        feature_name: Optional[str] = None,
        feature_md5: Optional[bool] = None,
    ) -> pd.DataFrame:
        """Read feature into a pandas DataFrame.

        Parameters
        ----------
        feature_name : str, optional
            Name of the feature to read (default None).
        feature_md5 : str, optional
            MD5 hash of the feature to read (default None).

        Returns
        -------
        pandas.DataFrame
            The features as a dataframe.

        """
        raise_error(
            msg="Concrete classes need to implement read_df().",
            klass=NotImplementedError,
        )

    @abstractmethod
    def store_metadata(self, meta_md5: str, element: Dict, meta: Dict) -> None:
        """Store metadata.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        meta : dict
            The metadata as a dictionary.
        """
        raise_error(
            msg="Concrete classes need to implement store_metadata().",
            klass=NotImplementedError,
        )

    def store(self, kind: str, **kwargs) -> None:
        """Store extracted features data.

        Parameters
        ----------
        kind : {"matrix", "timeseries", "table"}
            The storage kind.
        **kwargs
            The keyword arguments.

        Raises
        ------
        ValueError
            If ``kind`` is invalid.

        """
        # Do the check before calling the abstract methods, otherwise the
        # meta might be stored even if the data is not stored.
        if kind not in self._valid_inputs:
            raise_error(
                msg=f"I don't know how to store {kind}.",
                klass=ValueError,
            )
        t_meta = kwargs.pop("meta")
        meta_md5, t_meta, t_element = process_meta(t_meta)
        self.store_metadata(meta_md5=meta_md5, element=t_element, meta=t_meta)
        if kind == "matrix":
            self.store_matrix(meta_md5=meta_md5, element=t_element, **kwargs)
        elif kind == "timeseries":
            self.store_timeseries(
                meta_md5=meta_md5, element=t_element, **kwargs
            )
        elif kind == "table":
            self.store_table(meta_md5=meta_md5, element=t_element, **kwargs)

    def store_matrix(
        self,
        meta_md5: str,
        element: Dict,
        data: np.ndarray,
        col_names: Optional[Iterable[str]] = None,
        row_names: Optional[Iterable[str]] = None,
        matrix_kind: Optional[str] = "full",
        diagonal: bool = True,
    ) -> None:
        """Store matrix.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray
            The matrix data to store.
        col_names : list or tuple of str, optional
            The column names (default None).
        row_names : str, optional
            The column name to use in case number of rows greater than 1.
            If None and number of rows greater than 1, then the name will be
            "index" (default None).
        matrix_kind : str, optional
            The kind of matrix:

            * ``triu`` : store upper triangular only
            * ``tril`` : store lower triangular
            * ``full`` : full matrix

            (default "full").
        diagonal : bool, optional
            Whether to store the diagonal. If `matrix_kind` is "full", setting
            this to False will raise an error (default True).
        """
        raise_error(
            msg="Concrete classes need to implement store_matrix2d().",
            klass=NotImplementedError,
        )

    def store_table(
        self,
        meta_md5: str,
        element: Dict,
        data: Union[np.ndarray, List],
        columns: Optional[Iterable[str]] = None,
        rows_col_name: Optional[str] = None,
    ) -> None:
        """Store table.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray or list
            The table data to store.
        columns : list or tuple of str, optional
            The columns (default None).
        rows_col_name : str, optional
            The column name to use in case number of rows greater than 1.
            If None and number of rows greater than 1, then the name will be
            "index" (default None).
        """
        raise_error(
            msg="Concrete classes need to implement store_table().",
            klass=NotImplementedError,
        )

    def store_timeseries(
        self,
        meta_md5: str,
        element: Dict,
        data: np.ndarray,
        columns: Optional[Iterable[str]] = None,
    ) -> None:
        """Implement timeseries storing.

        Parameters
        ----------
        meta_md5 : str
            The metadata MD5 hash.
        element : dict
            The element as a dictionary.
        data : numpy.ndarray
            The timeseries data to store.
        columns : list or tuple of str, optional
            The column labels (default None).
        """
        raise_error(
            msg="Concrete classes need to implement store_timeseries().",
            klass=NotImplementedError,
        )

    @abstractmethod
    def collect(self) -> None:
        """Collect data."""
        raise_error(
            msg="Concrete classes need to implement collect().",
            klass=NotImplementedError,
        )

    def __str__(self) -> str:
        """Represent object as string.

        Returns
        -------
        str
            The string representation.

        """
        single = (
            "(single output)"
            if self.single_output is True
            else "(multiple output)"
        )
        return f"<{self.__class__.__name__} @ {self.uri} {single}>"



"""Provide imports for datagrabbers sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from .aomic_id1000_vbm import JuselessDataladAOMICID1000VBM
from .camcan_vbm import JuselessDataladCamCANVBM
from .ixi_vbm import JuselessDataladIXIVBM
from .ucla import JuselessUCLA
from .ukb_vbm import JuselessDataladUKBVBM



"""Provide tests for pattern."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path

import pytest

from junifer.datagrabber.pattern import PatternDataGrabber


def test_PatternDataGrabber_errors(tmp_path: Path) -> None:
    """Test PatternDataGrabber errors.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """

    with pytest.raises(TypeError, match=r"`types` must be a list"):
        PatternDataGrabber(
            datadir="/tmp",
            types="wrong",  # type: ignore
            patterns={"wrong": "pattern"},
            replacements="subject",  # type: ignore
        )

    with pytest.raises(TypeError, match=r"`types` must be a list of strings"):
        PatternDataGrabber(
            datadir="/tmp",  # type: ignore
            types=[1, 2, 3],  # type: ignore
            patterns={"1": "pattern", "2": "pattern", "3": "pattern"},
            replacements="subject",  # type: ignore
        )

    with pytest.raises(ValueError, match=r"must have the same length"):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns={"1": "pattern", "2": "pattern", "3": "pattern"},
            replacements=1,  # type: ignore
        )

    with pytest.raises(TypeError, match=r"`patterns` must be a dict"):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns="wrong",  # type: ignore
            replacements="subject",  # type: ignore
        )

    with pytest.raises(
        ValueError, match=r"`patterns` must have the same length"
    ):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns={"wrong": "pattern"},
            replacements="subject",  # type: ignore
        )

    with pytest.raises(
        ValueError, match=r"`patterns` must contain all `types`"
    ):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns={"wrong": "pattern", "func": "pattern"},
            replacements="subject",  # type: ignore
        )

    with pytest.raises(TypeError, match=r"must be a list of strings"):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns={"func": "func/test", "anat": "anat/test"},
            replacements=1,  # type: ignore
        )

    with pytest.raises(ValueError, match=r"not part of any pattern"):
        PatternDataGrabber(
            datadir="/tmp",
            types=["func", "anat"],
            patterns={
                "func": "func/{subject}.nii",
                "anat": "anat/{subject}.nii",
            },
            replacements=["subject", "wrong"],
        )

    tmpdir = tmp_path / "pattern_dg_test_errors"

    datagrabber = PatternDataGrabber(
        datadir=tmpdir,
        types=["func", "anat"],
        patterns={
            "func": "func/{subject}_single.nii",
            "anat": "anat/{subject}_{session}_ses.nii",
        },
        replacements=["subject", "session"],
    )

    with pytest.raises(ValueError, match="element keys must be"):
        datagrabber["sub001"]

    # This should not work, file does not exists
    with pytest.raises(ValueError, match="Cannot access"):
        datagrabber["sub001", "ses001"]

    (tmpdir / "func").mkdir(exist_ok=True, parents=True)
    (tmpdir / "anat").mkdir(exist_ok=True, parents=True)
    for t_subject in range(3):
        for t_session in range(2):
            subject = f"sub{t_subject:03d}"
            session = f"ses{t_session:03d}"
            (tmpdir / "func" / f"{subject}_single.nii").touch()
            if t_subject == 2:
                (tmpdir / "func" / f"{subject}_extra.nii").touch()
            (tmpdir / "anat" / f"{subject}_{session}_ses.nii").touch()

    # This should work, file now exists
    datagrabber["sub001", "ses001"]

    datagrabber = PatternDataGrabber(
        datadir=tmpdir,
        types=["func", "anat"],
        patterns={
            "func": "func/{subject}_*.nii",
            "anat": "anat/{subject}_{session}_*.nii",
        },
        replacements=["subject", "session"],
    )

    # access a subject with a missing session
    with pytest.raises(ValueError, match="No file matches"):
        datagrabber["sub001", "ses004"]

    # access a subject with two matching files
    with pytest.raises(ValueError, match="More than one"):
        datagrabber["sub002", "ses001"]

    # access the right one
    datagrabber["sub001", "ses001"]

    datagrabber = PatternDataGrabber(
        datadir=tmpdir,
        types=["func", "anat2"],
        patterns={
            "func": "func/{subject}_single.nii",
            "anat2": "anat2/{subject}_{session}_ses.nii",
        },
        replacements=["subject", "session"],
    )
    assert len(datagrabber.get_elements()) == 0


def test_PatternDataGrabber(tmp_path: Path) -> None:
    """Test PatternDataGrabber.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """

    datagrabber = PatternDataGrabber(
        datadir="/tmp/data",
        types=["func", "anat"],
        patterns={"func": "func/{subject}.nii", "anat": "anat/{subject}.nii"},
        replacements="subject",
    )
    assert datagrabber.datadir == Path("/tmp/data")
    assert datagrabber.types == ["func", "anat"]
    assert datagrabber.replacements == ["subject"]

    datagrabber = PatternDataGrabber(
        datadir=Path("/tmp/data"),
        types=["func", "anat"],
        patterns={
            "func": "func/{subject}.nii",
            "anat": "anat/{subject}_{session}.nii",
        },
        replacements=["subject", "session"],
    )
    assert datagrabber.datadir == Path("/tmp/data")
    assert datagrabber.types == ["func", "anat"]
    assert datagrabber.replacements == ["subject", "session"]

    tmpdir = tmp_path / "pattern_dg_test"
    (tmpdir / "func").mkdir(exist_ok=True, parents=True)
    (tmpdir / "anat").mkdir(exist_ok=True, parents=True)
    (tmpdir / "vbm").mkdir(exist_ok=True, parents=True)
    for t_subject in range(3):
        for t_session in range(2):
            for t_task in range(2, 4):
                subject = f"sub{t_subject:03d}"
                session = f"ses{t_session:03d}"
                task = f"task{t_task:03d}"
                if t_subject != 2:
                    (tmpdir / "func" / f"{subject}.nii").touch()
                (tmpdir / "anat" / f"{subject}_{session}.nii").touch()
                (tmpdir / "vbm" / f"{subject}_{task}_{session}.nii").touch()

    expected_elements = [
        ("sub000", "ses000"),
        ("sub000", "ses001"),
        ("sub001", "ses000"),
        ("sub001", "ses001"),
        ("sub002", "ses000"),
        ("sub002", "ses001"),
    ]

    datagrabber = PatternDataGrabber(
        datadir=tmpdir,
        types=["anat"],
        patterns={
            "anat": "anat/{subject}_{session}.nii",
        },
        replacements=["subject", "session"],
    )

    elements = datagrabber.get_elements()
    assert set(elements) == set(expected_elements)

    expected_elements = [
        ("sub000", "ses000", "task002"),
        ("sub000", "ses000", "task003"),
        ("sub000", "ses001", "task002"),
        ("sub000", "ses001", "task003"),
        ("sub001", "ses000", "task002"),
        ("sub001", "ses000", "task003"),
        ("sub001", "ses001", "task002"),
        ("sub001", "ses001", "task003"),
    ]

    datagrabber = PatternDataGrabber(
        datadir=tmpdir,
        types=["func", "anat", "vbm"],
        patterns={
            "func": "func/{subject}.nii",
            "anat": "anat/{subject}_{session}.nii",
            "vbm": "vbm/{subject}_{task}_{session}.nii",
        },
        replacements=["subject", "session", "task"],
    )

    elements = datagrabber.get_elements()
    assert set(elements) == set(expected_elements)

    out1 = datagrabber[("sub000", "ses000", "task002")]
    out2 = datagrabber[("sub000", "ses000", "task003")]

    assert out1["func"]["path"] == out2["func"]["path"]
    assert out1["anat"]["path"] == out2["anat"]["path"]
    assert out1["vbm"]["path"] != out2["vbm"]["path"]



"""Provide class for default data reader."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Dict, List, Optional

import nibabel as nib
import pandas as pd

from ..api.decorators import register_datareader
from ..pipeline import PipelineStepMixin, UpdateMetaMixin
from ..utils.logging import logger, warn_with_log


# Map each file extension to a type
_extensions = {
    ".nii": "NIFTI",
    ".nii.gz": "NIFTI",
    ".csv": "CSV",
    ".tsv": "TSV",
}

# Map each type to a function and arguments
_readers = {}
_readers["NIFTI"] = {"func": nib.load, "params": None}
_readers["CSV"] = {"func": pd.read_csv, "params": None}
_readers["TSV"] = {"func": pd.read_csv, "params": {"sep": "\t"}}


@register_datareader
class DefaultDataReader(PipelineStepMixin, UpdateMetaMixin):
    """Mixin class for default data reader."""

    def validate_input(self, input: List[str]) -> None:
        """Validate input.

        Parameters
        ----------
        input : list of str
            The input to the pipeline step. The list must contain the
            available Junifer Data dictionary keys.

        """
        # Nothing to validate, any input is fine
        pass

    def get_output_type(self, input: List[str]) -> List[str]:
        """Get output type.

        Parameters
        ----------
        input : list of str
            The input to the reader. The list must contain the
            available Junifer Data dictionary keys.

        Returns
        -------
        list of str
            The updated list of output types, as reading possibilities.

        """
        # It will output the same type of data as the input
        return input

    def fit_transform(
        self,
        input: Dict[str, Dict],
        params: Optional[Dict] = None,
    ) -> Dict:
        """Fit and transform.

        Parameters
        ----------
        input : dict
            The Junifer Data object.
        params : dict, optional
            Extra parameters for data types (default None).

        Returns
        -------
        dict
            The processed output as dictionary. The "data" key is added to
            each data type dictionary.

        """
        # For each type of data, try to read it
        out = input.copy()
        if params is None:
            params = {}
        for type_ in input.keys():
            if "path" not in input[type_]:
                warn_with_log(
                    f"Input type {type_} does not provide a path. Skipping."
                )
                continue
            t_path = input[type_]["path"]
            t_params = params.get(type_, {})

            # Convert to Path if datareader is not well done
            if not isinstance(t_path, Path):
                t_path = Path(t_path)
                out[type_]["path"] = t_path
            logger.info(f"Reading {type_} from {t_path.as_posix()}")
            fread = None

            fname = t_path.name.lower()
            for ext, ftype in _extensions.items():
                if fname.endswith(ext):
                    logger.info(f"{type_} is type {ftype}")
                    reader_func = _readers[ftype]["func"]
                    reader_params = _readers[ftype]["params"]
                    if reader_params is not None:
                        t_params.update(reader_params)
                    logger.debug(f"Calling {reader_func} with {t_params}")
                    fread = reader_func(t_path, **t_params)
                    break
            if fread is None:
                logger.info(
                    f"Unknown file type {t_path.as_posix()}, skipping reading"
                )
            out[type_]["data"] = fread
            self.update_meta(out[type_], "datareader")
        return out



"""
Generic BIDS datagrabber for datalad.
=====================================

This example uses a generic BIDS datagraber to get the data from a BIDS dataset
store in a datalad remote sibling.

Authors: Federico Raimondo

License: BSD 3 clause
"""

from junifer.datagrabber import PatternDataladDataGrabber
from junifer.utils import configure_logging


###############################################################################
# Set the logging level to info to see extra information
configure_logging(level="INFO")


###############################################################################
# The BIDS datagrabber requires three parameters: the types of data we want,
# the specific pattern that matches each type, and the variables that will be
# replaced int he patterns.
types = ["T1w", "BOLD"]
patterns = {
    "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
    "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
}
replacements = ["subject"]
###############################################################################
# Additionally, a datalad datagrabber requires the URI of the remote sibling
# and the location of the dataset within the remote sibling.
repo_uri = "https://gin.g-node.org/juaml/datalad-example-bids"
rootdir = "example_bids"

###############################################################################
# Now we can use the datagrabber within a `with` context
# One thing we can do with any datagrabber is iterate over the elements.
# In this case, each element of the datagrabber is one session.
with PatternDataladDataGrabber(
    rootdir=rootdir,
    types=types,
    patterns=patterns,
    uri=repo_uri,
    replacements=replacements,
) as dg:
    for elem in dg:
        print(elem)

###############################################################################
# Another feature of the datagrabber is the ability to get a specific
# element by its name. In this case, we index `sub-01` and we get the file
# paths for the two types of data we want (T1w and bold).
with PatternDataladDataGrabber(
    rootdir=rootdir,
    types=types,
    patterns=patterns,
    uri=repo_uri,
    replacements=replacements,
) as dg:
    sub01 = dg["sub-01"]
    print(sub01)



"""Provide mixin class for pipeline step."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from importlib.util import find_spec
from typing import Dict, List

from ..utils import raise_error


class PipelineStepMixin:
    """Mixin class for pipeline."""

    def validate_input(self, input: List[str]) -> None:
        """Validate the input to the pipeline step.

        Parameters
        ----------
        input : list of str
            The input to the pipeline step. The list must contain the
            available Junifer Data dictionary keys.

        Raises
        ------
        ValueError
            If the input does not have the required data.

        """
        raise_error(
            msg="Concrete classes need to implement validate_input().",
            klass=NotImplementedError,
        )

    def get_output_type(self, input_type: str) -> str:
        """Get output type.

        Parameters
        ----------
        input_type : str
            The data type input to the marker.

        Returns
        -------
        str
            The storage type output by the marker.

        """
        raise_error(
            msg="Concrete classes need to implement get_output_type().",
            klass=NotImplementedError,
        )

    def validate(self, input: List[str]) -> List[str]:
        """Validate the the pipeline step.

        Parameters
        ----------
        input : list of str
            The input to the pipeline step.

        Returns
        -------
        list of str
            The output of the pipeline step.

        Raises
        ------
        ValueError
            If the pipeline step object is missing dependencies required for
            its working or if the input does not have the required data.

        """
        # Check if _DEPENDENCIES attribute is found;
        # (markers and preprocessors will have them but not datareaders
        # as of now)
        dependencies_not_found = []
        if hasattr(self, "_DEPENDENCIES"):
            # Check if dependencies are importable
            for dependency in self._DEPENDENCIES:  # type: ignore
                if find_spec(dependency) is None:
                    dependencies_not_found.append(dependency)
        # Raise error if any dependency is not found
        if dependencies_not_found:
            raise_error(
                msg=f"{dependencies_not_found} are not installed but are "
                "required for using {self.name}.",
                klass=ImportError,
            )

        self.validate_input(input=input)
        outputs = [self.get_output_type(t_input) for t_input in input]
        return outputs

    def fit_transform(self, input: Dict[str, Dict]) -> Dict[str, Dict]:
        """Fit and transform.

        Parameters
        ----------
        input : list of str
            The input to the pipeline step. The list must contain the
            available Junifer Data dictionary keys.

        """
        raise_error(
            msg="Concrete classes need to implement fit_transform().",
            klass=NotImplementedError,
        )



"""Provide utilities for data module."""
from typing import List, Optional, Union

import numpy as np

from ..utils.logging import logger


def closest_resolution(
    resolution: Optional[float],
    valid_resolution: Union[List[float], List[int], np.ndarray],
) -> Union[float, int]:
    """Find the closest resolution.

    Parameters
    ----------
    resolution : float, optional
        The given resolution. If None, will return the highest resolution
        (default None).
    valid_resolution : list of float or int, or np.ndarray
        The array of valid resolutions.

    Returns
    -------
    float or int
        The closest valid resolution.
    """
    # Convert list of int to numpy.ndarray
    if not isinstance(valid_resolution, np.ndarray):
        valid_resolution = np.array(valid_resolution)

    if resolution is None:
        logger.info("Resolution set to None, using highest resolution.")
        closest = np.min(valid_resolution)
    elif any(x <= resolution for x in valid_resolution):
        # Case 1: get the highest closest resolution
        closest = np.max(valid_resolution[valid_resolution <= resolution])
    else:
        # Case 2: get the lower closest resolution
        closest = np.min(valid_resolution)

    return closest



"""Provide class and functions for logging."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import logging
import sys
from distutils.version import LooseVersion
from pathlib import Path
from subprocess import PIPE, Popen, TimeoutExpired
from typing import Dict, NoReturn, Optional, Type, Union
from warnings import warn

import datalad


logger = logging.getLogger("JUNIFER")

# Set up datalad logger level to warning by default
datalad.log.lgr.setLevel(logging.WARNING)

_logging_types = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


class WrapStdOut(logging.StreamHandler):
    """
    Dynamically wrap to sys.stdout.

    This makes packages that monkey-patch sys.stdout (e.g.doctest,
    sphinx-gallery) work properly.

    """

    def __getattr__(self, name: str) -> str:
        """Implement attribute fetch."""
        # Even more ridiculous than this class, this must be sys.stdout (not
        # just stdout) in order for this to work (tested on OSX and Linux)
        if hasattr(sys.stdout, name):
            return getattr(sys.stdout, name)
        else:
            raise AttributeError(f"'file' object has not attribute '{name}'")


def _get_git_head(path: Path) -> str:
    """Aux function to read HEAD from git.

    Parameters
    ----------
    path : pathlib.Path
        The path to read git HEAD from.

    Returns
    -------
    str
        Empty string if timeout expired for subprocess command execution else
        git HEAD information.

    """
    if not path.exists():
        raise ValueError(f"This path does not exist: {path}")
    command = f"cd {path}; git rev-parse --verify HEAD"
    process = Popen(
        args=command,
        stdout=PIPE,
        shell=True,
    )
    try:
        stdout, _ = process.communicate(timeout=10)
        proc_stdout = stdout.strip().decode()
    except TimeoutExpired:
        process.kill()
        proc_stdout = ""
    return proc_stdout


def get_versions() -> Dict:
    """Import stuff and get versions if module.

    Returns
    -------
    module_versions : dict
        The module names and corresponding versions.

    """
    module_versions = {}
    for name, module in sys.modules.items():
        if "." in name:
            continue
        if name in ["_curses"]:
            continue
        vstring = str(getattr(module, "__version__", None))
        module_version = LooseVersion(vstring)
        module_version = getattr(module_version, "vstring", None)
        if module_version is None:
            module_version = None
        elif "git" in module_version:
            git_path = Path(module.__file__).resolve().parent  # type: ignore
            head = _get_git_head(git_path)
            module_version += f"-HEAD:{head}"

        module_versions[name] = module_version
    return module_versions


# def get_ext_versions(tbox_path: Path) -> Dict:
#     """Get versions of external tools used by junifer.

#     Parameters
#     ----------
#     tbox_path : pathlib.Path
#         The path to external toolboxes.

#     Returns
#     -------
#     dict
#         The dependency information.

#     """
#     versions = {}
#     # spm_path = tbox_path / 'spm12'
#     # if spm_path.exists():
#     #     head = _get_git_head(spm_path)
#     #     module_version = 'SPM12-HEAD:{}'.format(head)
#     #     versions['spm'] = module_version
#     return versions


def _close_handlers(logger: logging.Logger) -> None:
    """Safely close relevant handlers for logger.

    Parameters
    ----------
    logger : logging.logger
        The logger to close handlers for.

    """
    for handler in list(logger.handlers):
        if isinstance(handler, (logging.FileHandler, logging.StreamHandler)):
            if isinstance(handler, logging.FileHandler):
                handler.close()
            logger.removeHandler(handler)


def _safe_log(versions: Dict, name: str) -> None:
    """Log with safety.

    Parameters
    ----------
    versions : dict
        The dictionary with keys as dependency names and values as the
        versions.
    name : str
        The dependency to look up in `versions`.

    """
    if name in versions:
        logger.info(f"{name}: {versions[name]}")


def log_versions(tbox_path: Optional[Path] = None) -> None:
    """Log versions of dependencies and junifer.

    If `tbox_path` is specified, can also log versions of external toolboxes.

    Parameters
    ----------
    tbox_path : pathlib.Path, optional
        The path to external toolboxes (default None).

    """
    # Get versions of all found packages
    versions = get_versions()

    logger.info("===== Lib Versions =====")
    _safe_log(versions, "numpy")
    _safe_log(versions, "scipy")
    _safe_log(versions, "pandas")
    _safe_log(versions, "nipype")
    _safe_log(versions, "nitime")
    _safe_log(versions, "nilearn")
    _safe_log(versions, "nibabel")
    _safe_log(versions, "junifer")
    logger.info("========================")

    if tbox_path is not None:
        # ext_versions = get_ext_versions(tbox_path)
        # logger.info('spm: {}'.format(ext_versions['spm']))
        # logger.info('========================')
        pass


def configure_logging(
    level: Union[int, str] = "WARNING",
    fname: Optional[Union[str, Path]] = None,
    overwrite: Optional[bool] = None,
    output_format=None,
) -> None:
    """Configure the logging functionality.

    Parameters
    ----------
    level : int or {"DEBUG", "INFO", "WARNING", "ERROR"}
        The level of the messages to print. If string, it will be interpreted
        as elements of logging (default "WARNING").
    fname : str or pathlib.Path, optional
        Filename of the log to print to. If None, stdout is used
        (default None).
    overwrite : bool, optional
        Overwrite the log file (if it exists). Otherwise, statements
        will be appended to the log (default). None is the same as False,
        but additionally raises a warning to notify the user that log
        entries will be appended (default None).
    output_format : str, optional
        Format of the output messages. See the following for examples:
        https://docs.python.org/dev/howto/logging.html
        e.g., ``"%(asctime)s - %(levelname)s - %(message)s"``.
        If None, default string format is used
        (default ``"%(asctime)s - %(name)s - %(levelname)s - %(message)s"``).

    """
    _close_handlers(logger)  # close relevant logger handlers

    # Set logging level
    if isinstance(level, str):
        level = _logging_types[level]

    # Set logging output handler
    if fname is not None:
        # Convert str to Path
        if not isinstance(fname, Path):
            fname = Path(fname)
        if fname.exists() and overwrite is None:
            warn(
                f"File ({str(fname.absolute())}) exists. "
                "Messages will be appended. Use overwrite=True to "
                "overwrite or overwrite=False to avoid this message."
            )
            overwrite = False
        mode = "w" if overwrite else "a"
        lh = logging.FileHandler(fname, mode=mode)
    else:
        lh = logging.StreamHandler(WrapStdOut())  # type: ignore

    # Set logging format
    if output_format is None:
        output_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        # (
        #     "%(asctime)s [%(levelname)s] %(message)s "
        #     "(%(filename)s:%(lineno)s)"
        # )
    formatter = logging.Formatter(fmt=output_format)

    lh.setFormatter(formatter)  # set formatter
    logger.setLevel(level)  # set level
    datalad.log.lgr.setLevel(level)  # set level for datalad
    logger.addHandler(lh)  # set handler
    log_versions()  # log versions of installed packages


def raise_error(
    msg: str,
    klass: Type[Exception] = ValueError,
    exception: Optional[Exception] = None,
) -> NoReturn:
    """Raise error, but first log it.

    Parameters
    ----------
    msg : str
        The message for the exception.
    klass : subclass of Exception, optional
        The subclass of Exception to raise using (default ValueError).
    exception : Exception, optional
        The original exception to follow up on (default None).

    """
    logger.error(msg)
    if exception is not None:
        raise klass(msg) from exception
    else:
        raise klass(msg)


def warn_with_log(
    msg: str, category: Optional[Type[Warning]] = RuntimeWarning
) -> None:
    """Warn, but first log it.

    Parameters
    ----------
    msg : str
        Warning message.
    category : subclass of Warning, optional
        The warning subclass (default RuntimeWarning).

    """
    logger.warning(msg)
    warn(msg, category=category)
